# BitBIRCH-Lean Python Package: An open-source clustering module based on iSIM.
#
# If you find this software useful please cite the following articles:
# - BitBIRCH: efficient clustering of large molecular libraries:
#   https://doi.org/10.1039/D5DD00030K
# - BitBIRCH Clustering Refinement Strategies:
#   https://doi.org/10.1021/acs.jcim.5c00627
# - BitBIRCH-Lean: TO-BE-ADDED
#
# Copyright (C) 2025  The Miranda-Quintana Lab and other BitBirch developers, comprised
# exclusively by:
# - Ramon Alain Miranda Quintana <ramirandaq@gmail.com>, <quintana@chem.ufl.edu>
# - Krisztina Zsigmond <kzsigmond@ufl.edu>
# - Ignacio Pickering <ipickering@chem.ufl.edu>
# - Kenneth Lopez Perez <klopezperez@chem.ufl.edu>
# - Miroslav Lzicar <miroslav.lzicar@deepmedchem.com>
#
# Authors of ./bblean/multiround.py are:
# - Ramon Alain Miranda Quintana <ramirandaq@gmail.com>, <quintana@chem.ufl.edu>
# - Ignacio Pickering <ipickering@chem.ufl.edu>
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# version 3 (SPDX-License-Identifier: GPL-3.0-only).
#
# Portions of this file are licensed under the BSD 3-Clause License
# Copyright (c) 2007-2024 The scikit-learn developers. All rights reserved.
# (SPDX-License-Identifier: BSD-3-Clause). Copies or reproductions of code in this
# file must in addition adhere to the BSD-3-Clause license terms. A
# copy of the BSD-3-Clause license can be located at the root of this repository, under
# ./LICENSES/BSD-3-Clause.txt.
#
# Portions of this file were previously licensed under the LGPL 3.0
# license (SPDX-License-Identifier: LGPL-3.0-only), they are relicensed in this program
# as GPL-3.0, with permission of all original copyright holders:
# - Ramon Alain Miranda Quintana <ramirandaq@gmail.com>, <quintana@chem.ufl.edu>
# - Vicky (Vic) Jung <jungvicky@ufl.edu>
# - Kenneth Lopez Perez <klopezperez@chem.ufl.edu>
# - Kate Huddleston <kdavis2@chem.ufl.edu>
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this
# program. This copy can be located at the root of this repository, under
# ./LICENSES/GPL-3.0-only.txt.  If not, see <http://www.gnu.org/licenses/gpl-3.0.html>.
r"""BitBirch 'Lean' class for fast, memory-efficient O(N) clustering"""
from __future__ import annotations  # Stringize type annotations for no runtime overhead
import itertools
import pickle
import sys
import typing_extensions as tpx
import os
import random
from pathlib import Path
import warnings
import typing as tp
from typing import cast
from collections import defaultdict
from weakref import WeakSet

import numpy as np
from numpy.typing import NDArray, DTypeLike

from bblean._memory import _mmap_file_and_madvise_sequential, _ArrayMemPagesManager
from bblean.merges import (
    MergeAcceptFunction,
    _get_merge_accept_fn,
    _BUILTIN_MERGES,
    DiscardSubcluster,
)
from bblean.utils import min_safe_uint
from bblean.fingerprints import (
    pack_fingerprints,
    _get_fingerprints_from_file_seq,
)
from bblean.similarity import (
    _jt_sim_arr_vec_packed,
    jt_most_dissimilar_packed,
    jt_isim_medoid,
    centroid_from_sum,
    estimate_jt_std,
    jt_isim,
)

if os.getenv("BITBIRCH_NO_EXTENSIONS"):
    from bblean.fingerprints import unpack_fingerprints as _unpack_fingerprints
else:
    try:
        # NOTE: There are small gains from using this fn but only ~3%, so don't warn for
        # now if this fails, and don't expose it
        from bblean._cpp_similarity import unpack_fingerprints as _unpack_fingerprints  # type: ignore # noqa
    except ImportError:
        from bblean.fingerprints import unpack_fingerprints as _unpack_fingerprints

__all__ = ["BitBirch"]


@tp.overload
def guess_threshold(
    fps: NDArray[np.uint8],
    input_is_packed: bool = True,
    n_features: int | None = None,
    max_samples: int = 1_000_000,
    factor: float = 3.0,
    return_mean_std: tp.Literal[False] = False,
) -> float:
    pass


@tp.overload
def guess_threshold(
    fps: NDArray[np.uint8],
    input_is_packed: bool = True,
    n_features: int | None = None,
    max_samples: int = 1_000_000,
    factor: float = 3.0,
    return_mean_std: tp.Literal[True] = True,
) -> tuple[float, float, float]:
    pass


def guess_threshold(
    fps: NDArray[np.uint8],
    input_is_packed: bool = True,
    n_features: int | None = None,
    max_samples: int = 1_000_000,
    factor: float = 3.0,
    return_mean_std: bool = False,
) -> float | tuple[float, float, float]:
    r""":meta private:

    Guess the optimal bitbirch threshold

    Uses the heuristic mean_tanimoto + 3.0 * std_tanimoto
    """
    num_fps = len(fps)
    if num_fps > max_samples:
        rng = np.random.default_rng(42)
        random_choices = rng.choice(num_fps, size=max_samples, replace=False)
        fps = fps[random_choices]
        num_fps = len(fps)
    mean = jt_isim(fps, input_is_packed, n_features)
    if num_fps <= 50:
        n_samples = num_fps
    else:
        n_samples = max(5 * np.sqrt(num_fps), 50)
    std = estimate_jt_std(
        fps, input_is_packed=input_is_packed, n_features=n_features, n_samples=n_samples
    )
    thresh = mean + factor * std
    if return_mean_std:
        return thresh, mean, std
    return thresh


# For backwards compatibility with the global "set_merge", keep weak references to all
# the BitBirch instances and update them when set_merge is called
_BITBIRCH_INSTANCES: WeakSet["BitBirch"] = WeakSet()


# For backwards compatibility: global function used to accept merges
_global_merge_accept: MergeAcceptFunction | None = None

_Input = tp.Union[NDArray[np.integer], list[NDArray[np.integer]]]


# For backwards compatibility: set the global merge_accept function
def set_merge(merge_criterion: str, tolerance: float = 0.05) -> None:
    r"""Sets the global criteria for merging subclusters in any BitBirch tree

    For usage see `BitBirch.set_merge`

    ..  warning::

        Use of this function is highly discouraged, instead use either:
        ``bb_tree = BitBirch(...)``
        ``bb_tree.set_merge(merge_criterion=..., tolerance=...)``
        or directly: ``bb_tree = BitBirch(..., merge_criterion=..., tolerance=...)``"

    """
    msg = (
        "Use of the global `set_merge` function is highly discouraged,\n"
        " instead use either: "
        " bb_tree = BitBirch(...)\n"
        " bb_tree.set_merge(merge_criterion=..., tolerance=...)\n"
        " or directly: `bb_tree = BitBirch(..., merge_criterion=..., tolerance=...)`."
    )
    warnings.warn(msg, UserWarning)
    # Set the global merge_accept function
    global _global_merge_accept
    _global_merge_accept = _get_merge_accept_fn(merge_criterion, tolerance)
    for bbirch in _BITBIRCH_INSTANCES:
        bbirch._merge_accept_fn = _global_merge_accept


# Utility function to validate the n_features argument for packed inputs
def _validate_n_features(
    X: _Input, input_is_packed: bool, n_features: int | None = None
) -> int:
    if len(X) == 0:
        raise ValueError("Input must have at least 1 fingerprint")
    if input_is_packed:
        _padded_n_features = len(X[0]) * 8 if isinstance(X, list) else X.shape[1] * 8
        if n_features is None:
            # Assume multiple of 8
            return _padded_n_features
        if _padded_n_features < n_features:
            raise ValueError(
                "n_features is larger than the padded length, which is inconsistent"
            )
        return n_features

    x_n_features = len(X[0]) if isinstance(X, list) else X.shape[1]
    if n_features is not None:
        if n_features != x_n_features:
            raise ValueError(
                "n_features is redundant for non-packed inputs"
                " if passed, it must be equal to X.shape[1] (or len(X[0]))."
                f" For passed X the inferred n_features was {x_n_features}."
                " If this value is not what you expected,"
                " make sure the passed X is actually unpacked."
            )
    return x_n_features


def _split_node(node: "_BFNode") -> tuple["_BFSubcluster", "_BFSubcluster"]:
    """The node has to be split if there is no place for a new subcluster
    in the node.
    1. An extra empty node and two empty subclusters are initialized.
    2. The pair of distant subclusters are found.
    3. The properties of the empty subclusters and nodes are updated
       according to the nearest distance between the subclusters to the
       pair of distant subclusters.
    4. The two nodes are set as children to the two subclusters.
    """
    n_features = node.n_features
    branching_factor = node.branching_factor
    new_subcluster1 = _BFSubcluster.empty(n_features)
    new_subcluster2 = _BFSubcluster.empty(n_features)

    node1 = _BFNode(branching_factor, n_features)
    node2 = node  # Rename for clarity
    new_subcluster1.child = node1
    new_subcluster2.child = node2

    if node2.is_leaf:
        # If is_leaf, _prev_leaf is guaranteed to be not None
        # NOTE: cast seems to have a small overhead here for some reason
        node1._prev_leaf = node2._prev_leaf
        node2._prev_leaf._next_leaf = node1  # type: ignore
        node1._next_leaf = node2
        node2._prev_leaf = node1

    # O(N) approximation to obtain "most dissimilar fingerprints" within an array
    node1_idx, _, node1_sim, node2_sim = jt_most_dissimilar_packed(
        node2.packed_centroids, n_features
    )
    node1_closer = node1_sim > node2_sim
    # Make sure node1 and node2 are closest to themselves, even if all sims are equal.
    # This can only happen when all node.packed_centroids are duplicates leading to all
    # distances between centroids being zero.

    # TODO: Currently this behavior is buggy (?), seems like in some cases one of the
    # subclusters may *never* get updated, double check this logic
    node1_closer[node1_idx] = True
    subclusters = node2._subclusters.copy()  # Shallow copy
    node2._subclusters = []  # Reset the node
    for idx, subcluster in enumerate(subclusters):
        if node1_closer[idx]:
            node1.append_subcluster(subcluster)
            new_subcluster1.update(subcluster)
        else:
            node2.append_subcluster(subcluster)
            new_subcluster2.update(subcluster)
    return new_subcluster1, new_subcluster2


class _BFNode:
    """Each node in a BitBirch tree is a _BFNode.

    The _BFNode holds a maximum of branching_factor _BFSubclusters.

    Parameters
    ----------
    branching_factor : int
        Maximum number of _BFSubcluster in the node.

    n_features : int
        The number of features.

    Attributes
    ----------
    _subclusters : list
        List of _BFSubcluster for thre _BFNode.

    _prev_leaf : _BFNode
        Only useful for leaf nodes, otherwise None

    _next_leaf : _BFNode
        Only useful for leaf nodes, otherwise None

    _packed_centroids_buf : NDArray[np.uint8]
        Packed array of shape (branching_factor + 1, (n_features + 7) // 8) The code
        internally manipulates this buf rather than packed_centroids, which is just a
        view of this buf.

    packed_centroids : ndarray of shape (branching_factor, n_features)
        Packed array of shape (len(_subclusters), (n_features + 7) // 8)
        View of the valid section of ``_packed_centroids_buf``.
    """

    # NOTE: Slots deactivates __dict__, and thus reduces memory usage of python objects
    __slots__ = (
        "n_features",
        "_subclusters",
        "_packed_centroids_buf",
        "_prev_leaf",
        "_next_leaf",
    )

    def __init__(self, branching_factor: int, n_features: int):
        self.n_features = n_features
        # The list of subclusters, centroids and squared norms
        # to manipulate throughout.
        self._subclusters: list["_BFSubcluster"] = []
        # Centroids are stored packed. All centroids up to branching_factor are
        # allocated in a contiguous array
        self._packed_centroids_buf = np.empty(
            (branching_factor + 1, (n_features + 7) // 8), dtype=np.uint8
        )
        # Nodes that are leaves have a non-null _prev_leaf
        self._prev_leaf: tp.Optional["_BFNode"] = None
        self._next_leaf: tp.Optional["_BFNode"] = None

    @property
    def is_leaf(self) -> bool:
        return self._prev_leaf is not None

    @property
    def branching_factor(self) -> int:
        return self._packed_centroids_buf.shape[0] - 1

    @property
    def packed_centroids(self) -> NDArray[np.uint8]:
        # packed_centroids returns a view of the valid part of _packed_centroids_buf.
        return self._packed_centroids_buf[: len(self._subclusters), :]

    def append_subcluster(self, subcluster: "_BFSubcluster") -> None:
        n_samples = len(self._subclusters)
        self._subclusters.append(subcluster)
        self._packed_centroids_buf[n_samples] = subcluster.packed_centroid

    def update_split_subclusters(
        self,
        subcluster: "_BFSubcluster",
        new_subcluster1: "_BFSubcluster",
        new_subcluster2: "_BFSubcluster",
    ) -> None:
        """Remove a subcluster from a node and update it with the
        split subclusters.
        """
        # Replace subcluster with new_subcluster1
        idx = self._subclusters.index(subcluster)
        self._subclusters[idx] = new_subcluster1
        self._packed_centroids_buf[idx] = new_subcluster1.packed_centroid
        # Append new_subcluster2
        self.append_subcluster(new_subcluster2)

    def insert_bf_subcluster(
        self,
        subcluster: "_BFSubcluster",
        merge_accept_fn: MergeAcceptFunction,
        threshold: float,
    ) -> bool:
        """Insert a new subcluster into the node."""
        if not self._subclusters:
            self.append_subcluster(subcluster)
            return False

        # Within this node, find the closest subcluster to the one to-be-inserted
        sim_matrix = _jt_sim_arr_vec_packed(
            self.packed_centroids, subcluster.packed_centroid
        )
        closest_idx = np.argmax(sim_matrix)
        closest_subcluster = self._subclusters[closest_idx]
        closest_node = closest_subcluster.child
        if closest_node is None:
            # The subcluster doesn't have a child node (this is a leaf node)
            # attempt direct merge
            merge_was_successful = closest_subcluster.merge_subcluster(
                subcluster, threshold, merge_accept_fn
            )
            if not merge_was_successful:
                # Could not merge due to criteria
                # Append subcluster, and check if splitting *this node* is needed
                self.append_subcluster(subcluster)
                return len(self._subclusters) > self.branching_factor
            # Merge success, update the centroid
            self._packed_centroids_buf[closest_idx] = closest_subcluster.packed_centroid
            return False

        # Hard case: the closest subcluster has a child (is 'tracking'), use recursion
        child_must_be_split = closest_node.insert_bf_subcluster(
            subcluster, merge_accept_fn, threshold
        )
        if child_must_be_split:
            # Split the child node and redistribute subclusters. Update
            # this node with the 'tracking' subclusters of the two new children.
            # Then, check if *this node* needs splitting too
            new_subcluster1, new_subcluster2 = _split_node(closest_node)
            self.update_split_subclusters(
                closest_subcluster, new_subcluster1, new_subcluster2
            )
            return len(self._subclusters) > self.branching_factor

        # Child need not be split, update the *tracking* closest subcluster
        closest_subcluster.update(subcluster)
        self._packed_centroids_buf[closest_idx] = self._subclusters[
            closest_idx
        ].packed_centroid
        return False


class _BFSubcluster:
    r"""Each subcluster in a BFNode is called a BFSubcluster.

    A BFSubcluster can have a BFNode as its child.

    Parameters
    ----------
    linear_sum : ndarray of shape (n_features,), default=None
        Sample. This is kept optional to allow initialization of empty
        subclusters.

    Attributes
    ----------
    n_samples : int
        Number of samples that belong to each subcluster.

    linear_sum : ndarray
        Linear sum of all the samples in a subcluster. Prevents holding
        all sample data in memory.

    packed_centroid : ndarray of shape (branching_factor + 1, n_features)
        Centroid of the subcluster. Prevent recomputing of centroids when
        ``BFNode.packed_centroids`` is called.

    mol_indices : list, default=[]
        List of indices of molecules included in the given cluster.

    child : _BFNode
        Child Node of the subcluster. Once a given _BFNode is set as the child
        of the _BFNode, it is set to ``self.child``.
    """

    # NOTE: Slots deactivates __dict__, and thus reduces memory usage of python objects
    __slots__ = ("_buffer", "packed_centroid", "child", "mol_indices")

    def __init__(
        self,
        buffer: NDArray[np.integer],
        mol_indices: tp.Sequence[int],
        packed_centroid: NDArray[np.uint8] | None = None,
        check_indices: bool = True,
    ) -> None:
        # If packed centroid is passed, it must be equal to the packed centroid
        # of the linear sum (this is not checked)
        if mol_indices and check_indices and buffer[-1] != len(mol_indices):
            raise ValueError("len mol_indices must be equal to buffer[-1] if specified")
        # NOTE: Internally, _buffer holds both "linear_sum" and "n_samples" It is
        # guaranteed to always have the minimum required uint dtype It should not be
        # accessed by external classes, only used internally. The individual parts can
        # be accessed in a read-only way using the linear_sum and n_samples
        # properties.
        #
        # IMPORTANT: To mutate instances of this class, *always* use the public API
        # given by replace|add_to_n_samples_and_linear_sum(...)
        self._buffer = buffer
        self.mol_indices = list(mol_indices)
        if packed_centroid is not None:
            self.packed_centroid = packed_centroid
        else:
            self.packed_centroid = centroid_from_sum(buffer[:-1], buffer[-1], pack=True)
        self.child: tp.Optional["_BFNode"] = None

    @classmethod
    def empty(cls, n_features: int) -> tpx.Self:
        packed_centroid = np.empty(0, dtype=np.uint8)  # Will be overwritten
        return cls(
            np.zeros((n_features + 1,), dtype=np.uint8),
            [],
            packed_centroid,
            check_indices=False,
        )

    @classmethod
    def from_fingerprint(
        cls, fp: NDArray[np.uint8], index: int, weight: int | None = None
    ) -> tpx.Self:
        if weight is not None:
            buffer = np.empty((len(fp) + 1,), dtype=min_safe_uint(weight))
            buffer[:-1] = fp
            buffer[-1] = 1
            # This inplace multiplication is actually safe, since we initialized
            # buffer with min_safe_uint, so it can be done without issue
            np.multiply(buffer, weight, out=buffer, casting="unsafe")
        else:
            buffer = np.empty((len(fp) + 1,), dtype=np.uint8)
            buffer[:-1] = fp
            buffer[-1] = 1
        packed_centroid = pack_fingerprints(fp)
        return cls(buffer, [index], packed_centroid, check_indices=False)

    @property
    def unpacked_centroid(self) -> NDArray[np.uint8]:
        return _unpack_fingerprints(self.packed_centroid, self.n_features)

    @property
    def n_features(self) -> int:
        return len(self._buffer) - 1

    @property
    def dtype_name(self) -> str:
        return self._buffer.dtype.name

    @property
    def linear_sum(self) -> NDArray[np.integer]:
        read_only_view = self._buffer[:-1]
        read_only_view.flags.writeable = False
        return read_only_view

    @property
    def n_samples(self) -> int:
        # Returns a python int, which is guaranteed to never overflow in sums, so
        # n_samples can always be safely added when accessed through this property
        return self._buffer.item(-1)

    # NOTE: Part of the contract is that all elements of linear sum must always be
    # less or equal to n_samples. This function does not check this
    def replace_n_samples_and_linear_sum(
        self, n_samples: int, linear_sum: NDArray[np.integer]
    ) -> None:
        # Cast to the minimum uint that can hold the inputs
        self._buffer = self._buffer.astype(min_safe_uint(n_samples), copy=False)
        # NOTE: Assignments are safe and do not recast the buffer
        self._buffer[:-1] = linear_sum
        self._buffer[-1] = n_samples
        self.packed_centroid = centroid_from_sum(linear_sum, n_samples, pack=True)

    # NOTE: Part of the contract is that all elements of linear sum must always be
    # less or equal to n_samples. This function does not check this
    def add_to_n_samples_and_linear_sum(
        self, n_samples: int, linear_sum: NDArray[np.integer]
    ) -> None:
        # Cast to the minimum uint that can hold the inputs
        new_n_samples = self.n_samples + n_samples
        self._buffer = self._buffer.astype(min_safe_uint(new_n_samples), copy=False)
        # NOTE: Assignment and inplace add are safe and do not recast the buffer
        self._buffer[:-1] += linear_sum
        self._buffer[-1] = new_n_samples
        self.packed_centroid = centroid_from_sum(
            self._buffer[:-1], new_n_samples, pack=True
        )

    def update(self, subcluster: "_BFSubcluster") -> None:
        self.add_to_n_samples_and_linear_sum(
            subcluster.n_samples, subcluster.linear_sum
        )
        self.mol_indices.extend(subcluster.mol_indices)

    def merge_subcluster(
        self,
        nominee_cluster: "_BFSubcluster",
        threshold: float,
        merge_accept_fn: MergeAcceptFunction,
    ) -> bool:
        """Check if a cluster is worthy enough to be merged. If yes, merge."""
        old_n = self.n_samples
        nom_n = nominee_cluster.n_samples
        new_n = old_n + nom_n
        old_ls = self.linear_sum
        nom_ls = nominee_cluster.linear_sum
        # np.add with explicit dtype is safe from overflows, e.g. :
        # np.add(np.uint8(255), np.uint8(255), dtype=np.uint16) = np.uint16(510)
        new_ls = np.add(old_ls, nom_ls, dtype=min_safe_uint(new_n))
        if merge_accept_fn(
            threshold,
            new_ls,
            new_n,
            old_ls,
            nom_ls,
            old_n,
            nom_n,
            self.mol_indices,
            nominee_cluster.mol_indices,
        ):
            self.replace_n_samples_and_linear_sum(new_n, new_ls)
            self.mol_indices.extend(nominee_cluster.mol_indices)
            return True
        return False


class _CentroidsMolIds(tp.TypedDict):
    centroids: list[NDArray[np.uint8]]
    mol_ids: list[list[int]]


class _MedoidsMolIds(tp.TypedDict):
    medoid_idxs: NDArray[np.int64]
    medoids: NDArray[np.uint8]
    mol_ids: list[list[int]]


class BitBirch:
    r"""Implements the BitBIRCH clustering algorithm, 'Lean' version

    Memory and time efficient, online-learning algorithm. It constructs a tree data
    structure with the cluster centroids being read off the leaf.

    If you find this software useful please cite the following articles:

    - BitBIRCH: efficient clustering of large molecular libraries:
      https://doi.org/10.1039/D5DD00030K
    - BitBIRCH Clustering Refinement Strategies:
      https://doi.org/10.1021/acs.jcim.5c00627
    - BitBIRCH-Lean: TO-BE-ADDED

    Parameters
    ----------

    threshold : float = 0.65
        The similarity radius of the subcluster obtained by merging a new sample and the
        closest subcluster should be greater than the threshold. Otherwise a new
        subcluster is started. Setting this value to be very low promotes splitting and
        vice-versa.

    branching_factor : int = 50
        Maximum number of 'BitFeatures' subclusters in each node. If a new sample enters
        such that the number of subclusters exceed the branching_factor then that node
        is split into two nodes with the subclusters redistributed in each. The parent
        subcluster of that node is removed and two new subclusters are added as parents
        of the 2 split nodes.

    merge_criterion: str
        radius, diameter or tolerance. *radius*: merge subcluster based on comparison to
        centroid of the cluster. *diameter*: merge subcluster based on instant Tanimoto
        similarity of cluster. *tolerance*: applies tolerance threshold to diameter
        merge criteria, which will merge subcluster with stricter threshold for newly
        added molecules.

    tolerance: float
        Penalty value for similarity threshold of the 'tolerance' merge criteria.

    Notes
    -----

    The tree data structure consists of nodes with each node holdint a number of
    subclusters (``BitFeatures``). The maximum number of subclusters in a node is
    determined by the branching factor. Each subcluster maintains a linear sum,
    mol_indices and the number of samples in that subcluster. In addition, each
    subcluster can also have a node as its child, if the subcluster is not a member of a
    leaf node.

    Each time a new fingerprint is fitted, it is merged with the subcluster closest to
    it and the linear sum, mol_indices and the number of samples int the corresponding
    subcluster are updated. This is done recursively untils the properties of a leaf
    node are updated.

    """

    def __init__(
        self,
        *,
        threshold: float = 0.65,
        branching_factor: int = 50,
        merge_criterion: str | MergeAcceptFunction | None = None,
        tolerance: float | None = None,
    ):
        # Criterion for merges
        self.threshold = threshold
        self.branching_factor = branching_factor
        if _global_merge_accept is not None:
            # Backwards compat
            if tolerance is not None:
                raise ValueError(
                    "tolerance can only be passed if "
                    "the *global* set_merge function has *not* been used"
                )
            if merge_criterion is not None:
                raise ValueError(
                    "merge_criterion can only be passed if "
                    "the *global* set_merge function has *not* been used"
                )
            self._merge_accept_fn = _global_merge_accept
        else:
            merge_criterion = "diameter" if merge_criterion is None else merge_criterion
            if isinstance(merge_criterion, MergeAcceptFunction):
                if tolerance is not None:
                    raise ValueError(
                        "'tolerance' arg is disregarded for custom merge functions"
                    )
                self._merge_accept_fn = merge_criterion
            else:
                tolerance = 0.05 if tolerance is None else tolerance
                self._merge_accept_fn = _get_merge_accept_fn(merge_criterion, tolerance)

        # Tree state
        self._num_fitted_fps = 0
        self._root: _BFNode | None = None
        self._dummy_leaf = _BFNode(branching_factor=2, n_features=0)
        # TODO: Type correctly
        self._global_clustering_centroid_labels: NDArray[np.int64] | None = None
        self._n_global_clusters = 0
        self._has_discarded = False

        # For backwards compatibility, weak-register in global state This is used to
        # update the merge_accept function if the global set_merge() is called
        # (discouraged)
        _BITBIRCH_INSTANCES.add(self)

    @property
    def merge_criterion(self) -> str:
        return self._merge_accept_fn.name

    @merge_criterion.setter
    def merge_criterion(self, value: str) -> None:
        self.set_merge(merge_criterion=value)

    @property
    def tolerance(self) -> float | None:
        fn = self._merge_accept_fn
        if hasattr(fn, "tolerance"):
            return fn.tolerance
        return None

    @tolerance.setter
    def tolerance(self, value: float) -> None:
        self.set_merge(tolerance=value)

    @property
    def is_init(self) -> bool:
        r"""Whether the tree has been initialized (True after first call to `fit()`)"""
        return self._dummy_leaf._next_leaf is not None

    @property
    def num_fitted_fps(self) -> int:
        r"""Total number of fitted fingerprints"""
        return self._num_fitted_fps

    def set_merge(
        self,
        merge_criterion: str | MergeAcceptFunction | None = None,
        *,
        tolerance: float | None = None,
        threshold: float | None = None,
        branching_factor: int | None = None,
    ) -> None:
        r"""Changes the criteria for merging subclusters in this BitBirch tree

        For an explanation of the parameters see the `BitBirch` class docstring.
        """
        if _global_merge_accept is not None:
            raise ValueError(
                "BitBirch.set_merge() can only called if "
                "the global set_merge() function has *not* been used"
            )
        _tolerance = 0.05 if tolerance is None else tolerance
        if isinstance(merge_criterion, MergeAcceptFunction):
            self._merge_accept_fn = merge_criterion
        elif isinstance(merge_criterion, str):
            self._merge_accept_fn = _get_merge_accept_fn(merge_criterion, _tolerance)
        if hasattr(self._merge_accept_fn, "tolerance"):
            self._merge_accept_fn.tolerance = _tolerance
        elif tolerance is not None:
            raise ValueError(f"Can't set tolerance for {self._merge_accept_fn}")
        if threshold is not None:
            self.threshold = threshold
        if branching_factor is not None:
            self.branching_factor = branching_factor

    def fit(
        self,
        X: _Input | Path | str,
        /,
        reinsert_indices: tp.Iterable[int] | None = None,
        input_is_packed: bool = True,
        n_features: int | None = None,
        max_fps: int | None = None,
        weights: tp.Iterable[int] | None = None,
    ) -> tpx.Self:
        r"""Build a BF Tree for the input data.

        Parameters
        ----------

        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Input data.

        reinsert_indices: Iterable[int]
            if ``reinsert_indices`` is passed, ``X`` corresponds only to the molecules
            that will be reinserted into the tree, and ``reinsert_indices`` are the
            indices associated with these molecules.

        input_is_packed: bool
            Whether the input fingerprints are packed

        n_features: int
            Number of featurs of input fingerprints. Only required for packed inputs if
            it is not a multiple of 8, otherwise it is redundant.

        Returns
        -------

        self
            Fitted estimator.
        """
        if isinstance(X, (Path, str)):
            X = _mmap_file_and_madvise_sequential(Path(X), max_fps=max_fps)
            mmanager = _ArrayMemPagesManager.from_bb_input(X)
        else:
            X = X[:max_fps]
            mmanager = _ArrayMemPagesManager.from_bb_input(X, can_release=False)

        n_features = _validate_n_features(X, input_is_packed, n_features)
        # Start a new tree the first time this function is called
        if self._only_has_leaves:
            raise ValueError("Internal nodes were released, call reset() before fit()")
        if not self.is_init:
            self._initialize_tree(n_features)
        self._root = cast("_BFNode", self._root)  # After init, this is not None

        # The array iterator either copies, un-sparsifies, or does nothing
        # with the array rows, depending on the kind of X passed
        arr_iterable = _get_array_iterable(X, input_is_packed, n_features)
        arr_iterable = cast(tp.Iterable[NDArray[np.uint8]], arr_iterable)
        iterable: tp.Iterable[tuple[int, NDArray[np.uint8]]]
        if reinsert_indices is None:
            iterable = enumerate(arr_iterable, self.num_fitted_fps)
        else:
            iterable = zip(reinsert_indices, arr_iterable)

        it_weights: tp.Iterator[int | None]
        if weights is None:
            it_weights = itertools.repeat(None)
        else:
            it_weights = iter(weights)

        threshold = self.threshold
        branching_factor = self.branching_factor
        merge_accept_fn = self._merge_accept_fn

        arr_idx = 0
        for idx, fp in iterable:
            subcluster = _BFSubcluster.from_fingerprint(fp, idx, next(it_weights))
            try:
                split = self._root.insert_bf_subcluster(
                    subcluster, merge_accept_fn, threshold
                )
            except DiscardSubcluster:
                self._has_discarded = True
                self._num_fitted_fps += 1
                arr_idx += 1
                continue

            if split:
                new_subcluster1, new_subcluster2 = _split_node(self._root)
                self._root = _BFNode(branching_factor, n_features)
                self._root.append_subcluster(new_subcluster1)
                self._root.append_subcluster(new_subcluster2)

            self._num_fitted_fps += 1
            arr_idx += 1
            if mmanager.can_release and mmanager.should_release_curr_page(arr_idx):
                mmanager.release_curr_page_and_update_addr()
        return self

    def _fit_buffers(
        self,
        X: _Input | Path | str,
        reinsert_index_seqs: tp.Iterable[tp.Sequence[int]] | None,
        check_indices: bool = True,
    ) -> tpx.Self:
        r"""Build a BF Tree starting from buffers

        Buffers are arrays of the form:
            - buffer[0:-1] = linear_sum
            - buffer[-1] = n_samples
        X is either an array or a list of such buffers

        If `reinsert_index_seqs` is passed, X corresponds only to the buffers to be
        reinserted into the tree, and `reinsert_index_seqs` are the sequences
        of indices associated with such buffers.

        If `reinsert_index_seqs` is None, then no indices are collected in the tree.
        Num samples is mutually exclusive with reinsert_index_seqs.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples + 1, n_features)
            Input data.

        Returns
        -------
        self
            Fitted estimator.
        """
        if isinstance(X, (Path, str)):
            X = _mmap_file_and_madvise_sequential(Path(X))
            mmanager = _ArrayMemPagesManager.from_bb_input(X)
        else:
            mmanager = _ArrayMemPagesManager.from_bb_input(X, can_release=False)

        n_features = _validate_n_features(X, input_is_packed=False) - 1
        # Start a new tree the first time this function is called
        if self._only_has_leaves:
            raise ValueError("Internal nodes were released, call reset() before fit()")
        if not self.is_init:
            self._initialize_tree(n_features)
        self._root = cast("_BFNode", self._root)  # After init, this is not None

        # The array iterator either copies, un-sparsifies, or does nothing with the
        # array rows, depending on the kind of X passed
        arr_iterable = _get_array_iterable(X, input_is_packed=False, dtype=X[0].dtype)
        merge_accept_fn = self._merge_accept_fn
        threshold = self.threshold
        branching_factor = self.branching_factor
        idx_provider: tp.Iterable[tp.Sequence[int]]
        arr_idx = 0
        if reinsert_index_seqs is None:
            idx_provider = itertools.repeat(())
        else:
            idx_provider = reinsert_index_seqs

        for idxs, buf in zip(idx_provider, arr_iterable):
            subcluster = _BFSubcluster(buf, idxs, check_indices=check_indices)
            try:
                split = self._root.insert_bf_subcluster(
                    subcluster, merge_accept_fn, threshold
                )
            except DiscardSubcluster:
                self._has_discarded = True
                self._num_fitted_fps += len(idxs)
                arr_idx += 1
                continue

            if split:
                new_subcluster1, new_subcluster2 = _split_node(self._root)
                self._root = _BFNode(branching_factor, n_features)
                self._root.append_subcluster(new_subcluster1)
                self._root.append_subcluster(new_subcluster2)

            self._num_fitted_fps += len(idxs)
            arr_idx += 1
            if mmanager.can_release and mmanager.should_release_curr_page(arr_idx):
                mmanager.release_curr_page_and_update_addr()
        return self

    # Provided for backwards compatibility
    def fit_reinsert(
        self,
        X: _Input | Path | str,
        reinsert_indices: tp.Iterable[int],
        input_is_packed: bool = True,
        n_features: int | None = None,
        max_fps: int | None = None,
    ) -> tpx.Self:
        r""":meta private:"""
        return self.fit(X, reinsert_indices, input_is_packed, n_features, max_fps)

    def _initialize_tree(self, n_features: int) -> None:
        # Initialize the root (and a dummy node to get back the subclusters
        self._root = _BFNode(self.branching_factor, n_features)
        self._dummy_leaf._next_leaf = self._root
        self._root._prev_leaf = self._dummy_leaf

    def _get_leaves(self) -> tp.Iterator[_BFNode]:
        r"""Yields all leaf nodes"""
        if not self.is_init:
            raise ValueError("The model has not been fitted yet.")
        leaf = self._dummy_leaf._next_leaf
        while leaf is not None:
            yield leaf
            leaf = leaf._next_leaf

    def get_centroids_mol_ids(
        self, sort: bool = True, packed: bool = True
    ) -> _CentroidsMolIds:
        """Get a dict with centroids and mol indices of the leaves"""
        # NOTE: This is different from the original bitbirch, here outputs are sorted by
        # default
        centroids = []
        mol_ids = []
        attr = "packed_centroid" if packed else "unpacked_centroid"
        for subcluster in self._get_leaf_bfs(sort=sort):
            centroids.append(getattr(subcluster, attr))
            mol_ids.append(subcluster.mol_indices)
        return {"centroids": centroids, "mol_ids": mol_ids}

    def get_centroids(
        self,
        sort: bool = True,
        packed: bool = True,
    ) -> list[NDArray[np.uint8]]:
        r"""Get a list of arrays with the centroids' fingerprints"""
        # NOTE: This is different from the original bitbirch, here outputs are sorted by
        # default
        attr = "packed_centroid" if packed else "unpacked_centroid"
        return [getattr(s, attr) for s in self._get_leaf_bfs(sort=sort)]

    def get_medoids_mol_ids(
        self,
        fps: NDArray[np.uint8],
        sort: bool = True,
        pack: bool = True,
        global_clusters: bool = False,
        input_is_packed: bool = True,
        n_features: int | None = None,
        iterative_unpacking: bool = False,
    ) -> _MedoidsMolIds:
        r"""Get a dict with medoid idxs, medoids and mol indices of the leaves

        The medoid indices are indices into the cluster mol ids, not into the fps array
        """
        cluster_members = self.get_cluster_mol_ids(
            sort=sort, global_clusters=global_clusters
        )

        if input_is_packed and not iterative_unpacking:
            fps = _unpack_fingerprints(fps, n_features=n_features)
            input_is_packed = False
        cluster_medoid_idxs, cluster_medoids = self._unpacked_medoids_from_members(
            fps, cluster_members, input_is_packed, n_features
        )
        if pack:
            cluster_medoids = pack_fingerprints(cluster_medoids)
        return {
            "medoid_idxs": cluster_medoid_idxs,
            "medoids": cluster_medoids,
            "mol_ids": cluster_members,
        }

    @staticmethod
    def _unpacked_medoids_from_members(
        fps: NDArray[np.uint8],
        cluster_members: tp.Sequence[list[int]],
        input_is_packed: bool = True,
        n_features: int | None = None,
    ) -> tuple[NDArray[np.int64], NDArray[np.uint8]]:
        if not input_is_packed:
            feats_num = fps.shape[1]
            n_features = None
        elif n_features is None:
            feats_num = fps.shape[1] * 8
        else:
            feats_num = n_features

        cluster_medoids = np.zeros((len(cluster_members), feats_num), dtype=np.uint8)
        cluster_medoid_idxs = np.zeros((len(cluster_members),), dtype=np.int64)
        for idx, members in enumerate(cluster_members):
            cluster_medoid_idxs[idx], cluster_medoids[idx, :] = jt_isim_medoid(
                fps[members],
                input_is_packed=input_is_packed,
                n_features=n_features,
                pack=False,
            )
        return cluster_medoid_idxs, cluster_medoids

    def get_medoids(
        self,
        fps: NDArray[np.uint8],
        sort: bool = True,
        pack: bool = True,
        global_clusters: bool = False,
        input_is_packed: bool = True,
        n_features: int | None = None,
    ) -> NDArray[np.uint8]:
        return self.get_medoids_mol_ids(
            fps, sort, pack, global_clusters, input_is_packed, n_features
        )["medoids"]

    def get_cluster_mol_ids(
        self, sort: bool = True, global_clusters: bool = False
    ) -> list[list[int]]:
        r"""Get the indices of the molecules in each cluster"""
        if global_clusters:
            if self._global_clustering_centroid_labels is None:
                raise ValueError(
                    "Must perform global clustering before fetching global labels"
                )
            bf_labels = (
                self._global_clustering_centroid_labels - 1
            )  # sub 1 to use as idxs

            # Collect the members of all clusters
            it = (bf.mol_indices for bf in self._get_leaf_bfs(sort=sort))
            return self._new_ids_from_labels(it, bf_labels, self._n_global_clusters)

        return [s.mol_indices for s in self._get_leaf_bfs(sort=sort)]

    @staticmethod
    def _new_ids_from_labels(
        members: tp.Iterable[list[int]],
        labels: NDArray[np.int64],
        n_labels: int | None = None,
    ) -> list[list[int]]:
        r"""Get the indices of the molecules in each cluster"""
        if n_labels is None:
            n_labels = len(np.unique(labels))
        new_members: list[list[int]] = [[] for _ in range(n_labels)]
        for i, idxs in enumerate(members):
            new_members[labels[i]].extend(idxs)
        return new_members

    def get_assignments(
        self,
        n_mols: int | None = None,
        sort: bool = True,
        check_valid: bool = True,
        global_clusters: bool = False,
    ) -> NDArray[np.uint64]:
        r"""Get an array with the cluster labels associated with each fingerprint idx"""
        if n_mols is not None:
            warnings.warn("The n_mols argument is redundant", DeprecationWarning)
        if n_mols is not None and n_mols != self.num_fitted_fps:
            raise ValueError(
                f"Provided n_mols {n_mols} is different"
                f" from the number of fitted fingerprints {self.num_fitted_fps}"
            )
        if check_valid or self._has_discarded:
            assignments = np.full(self.num_fitted_fps, 0, dtype=np.uint64)
        else:
            assignments = np.empty(self.num_fitted_fps, dtype=np.uint64)

        iterator: tp.Iterable[list[int]]
        if sort:
            iterator = self.get_cluster_mol_ids(sort=True)
        else:
            iterator = (
                s.mol_indices for leaf in self._get_leaves() for s in leaf._subclusters
            )

        if global_clusters:
            if self._global_clustering_centroid_labels is None:
                raise ValueError(
                    "Must perform global clustering before fetching global labels"
                )
            # Assign according to global clustering labels
            final_labels = self._global_clustering_centroid_labels
            for mol_ids, label in zip(iterator, final_labels):
                assignments[mol_ids] = label
        else:
            # Assign according to mol_ids from the subclusters
            for i, mol_ids in enumerate(iterator, 1):
                assignments[mol_ids] = i

        # Check that there are no unassigned molecules
        if not self._has_discarded and check_valid and (assignments == 0).any():
            raise ValueError("There are unasigned molecules")
        return assignments

    def dump_assignments(
        self,
        path: Path | str,
        smiles: tp.Iterable[str] = (),
        sort: bool = True,
        global_clusters: bool = False,
        check_valid: bool = True,
    ) -> None:
        r"""Dump the cluster assignments to a ``*.csv`` file"""
        import pandas as pd  # Hide pandas import since it is heavy

        path = Path(path)
        if isinstance(smiles, str):
            smiles = [smiles]
        smiles = np.asarray(smiles, dtype=np.str_)
        # Dump cluster assignments to *.csv
        assignments = self.get_assignments(
            sort=sort, check_valid=check_valid, global_clusters=global_clusters
        )
        if smiles.size and (len(assignments) != len(smiles)):
            raise ValueError(
                f"Len of the provided smiles {len(smiles)}"
                f" must match the number of fitted fingerprints {self.num_fitted_fps}"
            )
        df = pd.DataFrame({"assignments": assignments})
        if smiles.size:
            df["smiles"] = smiles
        df.to_csv(path, index=False)

    def reset(self) -> None:
        r"""Reset the tree state

        Delete *all internal nodes and leafs*, does not reset the merge criterion or
        other merge parameters.
        """
        # Reset the whole tree
        if self._root is not None:
            self._root._prev_leaf = None
            self._root._next_leaf = None
        self._dummy_leaf._next_leaf = None
        self._root = None
        self._num_fitted_fps = 0

    def delete_internal_nodes(self) -> None:
        r"""Delete all nodes in the tree that are not leaves

        This function is for advanced usage only. It should be called if there is need
        to use the BitBirch leaf clusters, but you need to release the memory held by
        the internal nodes. After calling this function, no more fingerprints can be fit
        into the tree, unless a call to `BitBirch.reset` afterwards releases the
        *whole tree*, including the leaf clusters.
        """
        if not tp.cast(_BFNode, self._root).is_leaf:
            # release all nodes that are not leaves,
            # they are kept alive by references from dummy_leaf
            self._root = None

    @property
    def _only_has_leaves(self) -> bool:
        return (self._root is None) and (self._dummy_leaf._next_leaf is not None)

    def recluster_inplace(
        self,
        iterations: int = 1,
        extra_threshold: float = 0.0,
        shuffle: bool = False,
        seed: int | None = None,
        verbose: bool = False,
        stop_early: bool = False,
    ) -> tpx.Self:
        r"""Refine singleton clusters by re-inserting them into the tree

        Parameters
        ----------
        extra_threshold : float, default=0.0
            The amount to increase the current threshold in each iteration.

        iterations : int, default=1
            The maximum number of refinement iterations to perform.

        Returns
        -------
        self : BitBirch
            The fitted estimator with refined clusters.

        Raises
        ------
        ValueError
            If the model has not been fitted.
        """
        if not self.is_init:
            raise ValueError("The model has not been fitted yet.")

        singletons_before = 0
        for _ in range(iterations):
            # Get the BFs
            bfs = self._get_leaf_bfs(sort=True)

            # Count the number of clusters and singletons
            singleton_bfs = sum(1 for bf in bfs if bf.n_samples == 1)

            # Check stopping criteria
            if stop_early:
                if singleton_bfs == 0 or singleton_bfs == singletons_before:
                    # No more singletons to refine
                    break
            singletons_before = singleton_bfs

            # Print progress
            if verbose:
                print(f"Current number of clusters: {len(bfs)}")
                print(f"Current number of singletons: {singleton_bfs}")

            if shuffle:
                random.seed(seed)
                random.shuffle(bfs)

            # Prepare the buffers for refitting
            fps_bfs, mols_bfs = self._prepare_bf_to_buffer_dicts(bfs)

            # Reset the tree
            self.reset()

            # Change the threshold
            self.threshold += extra_threshold

            # Refit the subsclusters
            for bufs, mol_idxs in zip(fps_bfs.values(), mols_bfs.values()):
                self._fit_buffers(bufs, reinsert_index_seqs=mol_idxs)

        # Print final stats
        if verbose:
            bfs = self._get_leaf_bfs(sort=True)
            singleton_bfs = sum(1 for bf in bfs if bf.n_samples == 1)
            print(f"Final number of clusters: {len(bfs)}")
            print(f"Final number of singletons: {singleton_bfs}")
        return self

    def refine_inplace(
        self,
        X: _Input | Path | str | tp.Sequence[Path],
        initial_mol: int = 0,
        input_is_packed: bool = True,
        n_largest: int = 1,
    ) -> tpx.Self:
        r"""Refine the tree: break the largest clusters in singletons and re-fit"""
        if not self.is_init:
            raise ValueError("The model has not been fitted yet.")
        # Release the memory held by non-leaf nodes
        self.delete_internal_nodes()

        # Extract the BitFeatures of the leaves, breaking the largest cluster apart into
        # singleton subclusters
        fps_bfs, mols_bfs = self._bf_to_np_refine(  # This function takes a bunch of mem
            X,
            initial_mol=initial_mol,
            input_is_packed=input_is_packed,
            n_largest=n_largest,
        )
        # Reset the tree
        self.reset()

        # Rebuild the tree again from scratch, reinserting all the subclusters
        for bufs, mol_idxs in zip(fps_bfs.values(), mols_bfs.values()):
            self._fit_buffers(bufs, reinsert_index_seqs=mol_idxs)
        return self

    def _get_leaf_bfs(self, sort: bool = True) -> list[_BFSubcluster]:
        r"""Get the BitFeatures of the leaves"""
        bfs = [s for leaf in self._get_leaves() for s in leaf._subclusters]
        if sort:
            # Sort the BitFeatures by the number of samples in the cluster
            bfs.sort(key=lambda x: x.n_samples, reverse=True)
        return bfs

    def _bf_to_np_refine(
        self,
        X: _Input | Path | str | tp.Sequence[Path],
        initial_mol: int = 0,
        input_is_packed: bool = True,
        n_largest: int = 1,
    ) -> tuple[dict[str, list[NDArray[np.integer]]], dict[str, list[list[int]]]]:
        """Prepare numpy bufs ('np') for BitFeatures, splitting the biggest n clusters

        The largest clusters are split into singletons. In order to perform this split,
        the *original* fingerprint array used to fit the tree (X) has to be provided,
        together with the index associated with the first fingerprint.

        The split is only performed for the returned 'np' buffers, the clusters in the
        tree itself are not modified
        """
        if n_largest == 0:
            return self._bf_to_np()

        if n_largest < 1:
            raise ValueError("n_largest must be >= 1")

        bfs = self._get_leaf_bfs()
        largest = bfs[:n_largest]
        rest = bfs[n_largest:]
        n_features = largest[0].n_features

        dtypes_to_fp, dtypes_to_mols = self._prepare_bf_to_buffer_dicts(rest)
        # Add X and mol indices of the "big" cluster
        if input_is_packed:
            unpack_or_copy = lambda x: _unpack_fingerprints(
                cast(NDArray[np.uint8], x), n_features
            )
        else:
            unpack_or_copy = lambda x: x.copy()

        for big_bf in largest:
            full_arr_idxs = [(idx - initial_mol) for idx in big_bf.mol_indices]
            _X: _Input
            if isinstance(X, (Path, str)):
                # Only load the specific required mol idxs
                _X = cast(NDArray[np.integer], np.load(X, mmap_mode="r"))[full_arr_idxs]
                arr_idxs = list(range(len(_X)))
                mol_idxs = big_bf.mol_indices
            elif isinstance(X[0], Path):
                # Only load the specific required mol idxs
                sort_idxs = np.argsort(full_arr_idxs)
                _X = _get_fingerprints_from_file_seq(
                    cast(tp.Sequence[Path], X),
                    [full_arr_idxs[i] for i in sort_idxs],
                )
                arr_idxs = list(range(len(_X)))
                mol_idxs = big_bf.mol_indices
                mol_idxs = [mol_idxs[i] for i in sort_idxs]
            else:
                # Index the full array / list
                _X = cast(_Input, X)
                arr_idxs = full_arr_idxs
                mol_idxs = big_bf.mol_indices

            for mol_idx, arr_idx in zip(mol_idxs, arr_idxs):
                buffer = np.empty(n_features + 1, dtype=np.uint8)
                buffer[:-1] = unpack_or_copy(_X[arr_idx])
                buffer[-1] = 1
                dtypes_to_fp["uint8"].append(buffer)
                dtypes_to_mols["uint8"].append([mol_idx])
        return dtypes_to_fp, dtypes_to_mols

    def _bf_to_np(
        self,
    ) -> tuple[dict[str, list[NDArray[np.integer]]], dict[str, list[list[int]]]]:
        """Prepare numpy buffers ('np') for BitFeatures of all clusters"""
        return self._prepare_bf_to_buffer_dicts(self._get_leaf_bfs())

    @staticmethod
    def _prepare_bf_to_buffer_dicts(
        bfs: list["_BFSubcluster"],
    ) -> tuple[dict[str, list[NDArray[np.integer]]], dict[str, list[list[int]]]]:
        # Helper function used when returning lists of subclusters
        dtypes_to_fp = defaultdict(list)
        dtypes_to_mols = defaultdict(list)
        for bf in bfs:
            dtypes_to_fp[bf.dtype_name].append(bf._buffer)
            dtypes_to_mols[bf.dtype_name].append(bf.mol_indices)
        return dtypes_to_fp, dtypes_to_mols

    def __repr__(self) -> str:
        fn = self._merge_accept_fn
        parts = [
            f"threshold={self.threshold}",
            f"branching_factor={self.branching_factor}",
            f"merge_criterion='{fn.name if fn.name in _BUILTIN_MERGES else fn}'",
        ]
        if self.tolerance is not None:
            parts.append(f"tolerance={self.tolerance}")
        return f"{self.__class__.__name__}({', '.join(parts)})"

    def save(self, path: Path | str) -> None:
        r""":meta private:"""
        # TODO: BitBIRCH is highly recursive. pickling may crash python,
        # an alternative solution would be better
        msg = (
            "Saving large BitBIRCH trees may result in large memory peaks."
            " An alternative serialization method may be implemented in the future"
        )
        warnings.warn(msg)
        _old_limit = sys.getrecursionlimit()
        sys.setrecursionlimit(1_000_000_000)
        with open(path, mode="wb") as f:
            pickle.dump(self, f)
        sys.setrecursionlimit(_old_limit)

    @classmethod
    def load(cls, path: Path | str) -> tpx.Self:
        r""":meta private:"""
        # TODO: BitBIRCH is highly recursive. pickling may crash python,
        # an alternative solution would be better
        msg = (
            "Loading large BitBIRCH trees may result in large memory peaks."
            " An alternative serialization method may be implemented in the future"
        )
        warnings.warn(msg)
        _old_limit = sys.getrecursionlimit()
        sys.setrecursionlimit(1_000_000_000)
        with open(path, mode="rb") as f:
            tree = pickle.load(f)
        sys.setrecursionlimit(_old_limit)
        if not isinstance(tree, cls):
            raise ValueError("Path does not contain a bitbirch object")
        return tree

    def global_clustering(
        self,
        n_clusters: int,
        *,
        method: str = "kmeans",
        # TODO: Type correctly
        **method_kwargs: tp.Any,
    ) -> tpx.Self:
        r""":meta private:"""
        warnings.warn(
            "Global clustering is an experimental features"
            " it will be modified without warning, please do not use"
        )
        if not self.is_init:
            raise ValueError("The model has not been fitted yet.")
        centroids = np.vstack(self.get_centroids(packed=False))
        labels = self._centrals_global_clustering(
            centroids, n_clusters, method=method, input_is_packed=False, **method_kwargs
        )
        num_centroids = len(centroids)
        self._n_global_clusters = (
            n_clusters if num_centroids > n_clusters else num_centroids
        )
        self._global_clustering_centroid_labels = labels
        return self

    @staticmethod
    def _centrals_global_clustering(
        centrals: NDArray[np.uint8],
        n_clusters: int,
        *,
        method: str = "kmeans",
        input_is_packed: bool = True,
        n_features: int | None = None,
        # TODO: Type correctly
        **method_kwargs: tp.Any,
    ) -> NDArray[np.int64]:
        r""":meta private:"""
        if method not in {"agglomerative", "kmeans", "kmeans-normalized"}:
            raise ValueError(f"Unknown method {method}")
        # Returns the labels associated with global clustering
        # Lazy import because sklearn is very heavy
        from sklearn.cluster import KMeans, AgglomerativeClustering
        from sklearn.exceptions import ConvergenceWarning

        if input_is_packed:
            centrals = _unpack_fingerprints(centrals, n_features)

        num_centrals = len(centrals)
        if num_centrals < n_clusters:
            msg = (
                f"Number of subclusters found ({num_centrals}) by BitBIRCH is less "
                "than ({n_clusters}). Decrease k or the threshold."
            )
            warnings.warn(msg, ConvergenceWarning, stacklevel=2)
            n_clusters = num_centrals

        if method == "kmeans-normalized":
            centrals = centrals / np.linalg.norm(centrals, axis=1, keepdims=True)
        if method in ["kmeans", "kmeans-normalized"]:
            predictor = KMeans(n_clusters=n_clusters, **method_kwargs)
        elif method == "agglomerative":
            predictor = AgglomerativeClustering(n_clusters=n_clusters, **method_kwargs)
        else:
            raise ValueError("method must be one of 'kmeans' or 'agglomerative'")

        # Add 1 to start labels from 1 instead of 0, so 0 can be used as sentinel
        # value
        # This is the bottleneck for building this index
        # K-means is feasible, agglomerative is extremely expensive
        return predictor.fit_predict(centrals) + 1


# There are 4 cases here:
# (1) The input is a scipy.sparse array
# (2) The input is a list of dense arrays (nothing required)
# (3) The input is a packed array or list of packed arrays (unpack required)
# (4) The input is a dense array (copy required)
# NOTE: Sparse iteration hack is taken from sklearn
# It returns a densified row when iterating over a sparse matrix, instead
# of constructing a sparse matrix for every row that is expensive.
#
# Output is *always* of dtype uint8, but input (if unpacked) can be of arbitrary dtype
# It is most efficient for input to be uint8 to prevent copies
def _get_array_iterable(
    X: _Input,
    input_is_packed: bool = True,
    n_features: int | None = None,
    dtype: DTypeLike = np.uint8,
) -> tp.Iterable[NDArray[np.integer]]:
    if input_is_packed:
        # Unpacking copies the fingerprints, so no extra copy required
        # NOTE: cast seems to have a very small overhead in this loop for some reason
        return (_unpack_fingerprints(a, n_features) for a in X)  # type: ignore
    if isinstance(X, list):
        # No copy is required here unless the dtype is not uint8
        return (a.astype(dtype, copy=False) for a in X)
    if isinstance(X, np.ndarray):
        # A copy is required here to avoid keeping a ref to the full array alive
        return (a.astype(dtype, copy=True) for a in X)
    return _iter_sparse(X)


# NOTE: In practice this branch is never used, it could probably safely be deleted
def _iter_sparse(X: tp.Any) -> tp.Iterator[NDArray[np.uint8]]:
    import scipy.sparse  # Hide this import since scipy is heavy

    if not scipy.sparse.issparse(X):
        raise ValueError(f"Input of type {type(X)} is not supported")
    n_samples, n_features = X.shape
    X_indices = X.indices
    X_data = X.data
    X_indptr = X.indptr
    for i in range(n_samples):
        a = np.zeros(n_features, dtype=np.uint8)
        startptr, endptr = X_indptr[i], X_indptr[i + 1]
        nonzero_indices = X_indices[startptr:endptr]
        a[nonzero_indices] = X_data[startptr:endptr].astype(np.uint8, copy=False)
        yield a



r"""Plotting and visualization convenience functions"""

import warnings
from pathlib import Path
import pickle
import random
import typing as tp

import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import seaborn as sns
from rdkit import Chem
from rdkit.Chem import Draw
import colorcet
from sklearn.preprocessing import StandardScaler, normalize as normalize_features
from sklearn.decomposition import PCA
from openTSNE.sklearn import TSNE
from openTSNE.affinity import Multiscale
import umap

from bblean.utils import batched, _num_avail_cpus, _has_files_or_valid_symlinks
from bblean.analysis import ClusterAnalysis, cluster_analysis
from bblean._config import TSNE_SEED

__all__ = [
    "summary_plot",
    "tsne_plot",
    "umap_plot",
    "pops_plot",
    "pca_plot",
    "dump_mol_images",
]


def isim_size_ordered_plot(
    c: ClusterAnalysis,
    /,
    threshold: float | None = None,
    title: str | None = None,
) -> tuple[plt.Figure, tuple[plt.Axes, ...]]:
    r"""iSIMs of clusters ordered by size"""
    fig, axs = plt.subplots(1, 2, figsize=(6.4, 3), sharex=True, sharey=True)

    # ordered by size by default, note that this generates some NaNs
    cluster_isims = c.isims
    seq_idxs = np.arange(1, len(cluster_isims) + 1)
    axs[0].set_ylim(0, 1.0)
    axs[0].scatter(seq_idxs, cluster_isims, color="tab:green", s=1.5, alpha=0.5)
    pad = 5
    kernel = np.ones(11) / 11
    padded = np.pad(cluster_isims, (pad, pad), mode="edge")
    moving_average = np.convolve(padded, kernel, mode="valid")
    moving_average_sq = np.convolve(
        padded**2,
        kernel,
        mode="valid",
    )
    moving_std = np.sqrt(moving_average_sq - moving_average**2)
    axs[1].plot(
        seq_idxs,
        moving_average,
        color="darkgreen",
        linewidth=1,
        linestyle="solid",
        label="Moving average",
    )
    axs[1].fill_between(
        seq_idxs,
        moving_average - moving_std,
        moving_average + moving_std,
        color="tab:gray",
        alpha=0.4,
        label="Moving std.",
    )
    axs[1].legend()
    axs[0].set_xlabel("Cluster label")
    axs[1].set_xlabel("Cluster label")
    axs[0].set_ylabel("Cluster iSIM")
    msg = f"iSIM for top {c.clusters_num} largest clusters, ordered by size"
    if threshold is not None:
        for ax in axs:
            ax.axhline(threshold, linestyle="dashed", linewidth=0.5, color="k")
            ax.text(
                0.02,
                threshold - 0.03,
                f"threshold = {threshold:.2f}",
                transform=ax.get_yaxis_transform(),  # x in axes, y in data coords
                fontsize=5,
            )
    if title is not None:
        msg = f"{msg} for {title}"
    fig.suptitle(msg)
    return fig, (axs[0], axs[1])


def isim_dist_plot(
    c: ClusterAnalysis,
    /,
    threshold: float | None = None,
    title: str | None = None,
) -> tuple[plt.Figure, tuple[plt.Axes, ...]]:
    r"""Distribution of cluster populations using KDE"""
    fig, ax = plt.subplots()
    cluster_isims = c.isims
    sns.kdeplot(
        ax=ax,
        data=cluster_isims,
        color="tab:green",
        bw_adjust=0.25,
        gridsize=len(cluster_isims) // 5,
        fill=True,
        warn_singular=False,
    )
    ax.set_ylabel("Density")
    ax.set_xlim(0, 1.0)
    ax.set_xlabel("Cluster iSIM")
    msg = f"iSIM distribution for top {c.clusters_num} largest clusters"
    if threshold is not None:
        ax.axvline(threshold, linestyle="dashed", linewidth=0.5, color="k")
        ax.text(
            threshold - 0.03,
            0.3,
            f"threshold = {threshold:.2f}",
            transform=ax.get_xaxis_transform(),  # x in axes, y in data coords
            rotation=90,
            fontsize=5,
        )
    if title is not None:
        msg = f"{msg} for {title}"
    fig.suptitle(msg)
    return fig, (ax,)


def pops_plot(
    c: ClusterAnalysis,
    /,
    title: str | None = None,
) -> tuple[plt.Figure, tuple[plt.Axes, ...]]:
    r"""Distribution of cluster populations using KDE"""
    fig, ax = plt.subplots()
    cluster_sizes = c.sizes
    sns.kdeplot(
        ax=ax,
        data=cluster_sizes,
        color="tab:purple",
        bw_adjust=0.25,
        gridsize=len(cluster_sizes) // 5,
        fill=True,
        warn_singular=False,
    )
    ax.set_ylabel("Density")
    ax.set_xlabel("Cluster size")
    msg = f"Populations for top {c.clusters_num} largest clusters"
    if c.min_size is not None:
        msg = f"{msg} (min. size = {c.min_size})"
    if title is not None:
        msg = f"{msg} for {title}"
    fig.suptitle(msg)
    return fig, (ax,)


# Similar to "init_plot" in the original bitbirch
def summary_plot(
    c: ClusterAnalysis,
    /,
    title: str | None = None,
    counts_ylim: int | None = None,
    annotate: bool = True,
) -> tuple[plt.Figure, tuple[plt.Axes, ...]]:
    r"""Create a summary plot from a cluster analysis

    If the analysis contains scaffolds, a scaffold analysis is added to the plot"""
    orange = "tab:orange"
    blue = "tab:blue"
    if mpl.rcParamsDefault["font.size"] == plt.rcParams["font.size"]:
        plt.rcParams["font.size"] = 8
    if annotate:
        fig, ax = plt.subplots(figsize=(5, 2.5), dpi=250, constrained_layout=True)
    else:
        fig, ax = plt.subplots()

    # Plot and annotate the number of molecules
    label_strs = c.labels.astype(str)  # TODO: Is this necessary?
    ax.bar(
        label_strs,
        c.sizes,
        color=blue,
        label="Num. molecules",
        zorder=0,
    )
    ax.set_ylim(0, counts_ylim)
    if annotate:
        for i, mol in enumerate(c.sizes):
            plt.text(
                i,
                mol,
                f"{mol}",
                ha="center",
                va="bottom",
                color="black",
                fontsize=5,
            )

    if c.has_scaffolds:
        # Plot and annotate the number of unique scaffolds
        plt.bar(
            label_strs,
            c.unique_scaffolds_num,
            color=orange,
            label="Num. unique scaffolds",
            zorder=1,
        )
        if annotate:
            for i, s in enumerate(c.unique_scaffolds_num):
                plt.text(
                    i,
                    s,
                    f"{s}",
                    ha="center",
                    va="bottom",
                    color="white",
                    fontsize=5,
                )

    # Labels
    ax.set_xlabel("Cluster label")
    ax.set_ylabel("Num. molecules")
    ax.set_xticks(range(c.clusters_num))

    # Plot iSIM
    if c.has_fps:
        ax_isim = ax.twinx()
        ax_isim.plot(
            c.labels - 1,
            c.isims,
            color="tab:green",
            linestyle="dashed",
            linewidth=1.5,
            zorder=5,
            alpha=0.6,
        )
        ax_isim.scatter(
            c.labels - 1,
            c.isims,
            color="tab:green",
            marker="o",
            s=15,
            label="Tanimoto iSIM",
            edgecolor="darkgreen",
            zorder=100,
            alpha=0.6,
        )
        ax_isim.set_ylabel("Tanimoto iSIM (average similarity)")
        ax_isim.set_yticks(np.arange(0, 1.1, 0.1))
        ax_isim.set_ylim(0, 1)
        ax_isim.spines["right"].set_color("tab:green")
        ax_isim.tick_params(colors="tab:green")
        ax_isim.yaxis.label.set_color("tab:green")
    bbox = ax.get_position()
    fig.legend(
        loc="upper right",
        bbox_to_anchor=(bbox.x0 + 0.95 * bbox.width, bbox.y0 + 0.95 * bbox.height),
    )
    if c.has_all_clusters:
        msg = "Metrics of all clusters"
    else:
        msg = f"Metrics of top {c.clusters_num} largest clusters"
    if title is not None:
        msg = f"{msg} for {title}"
    fig.suptitle(msg)
    if not c.has_fps:
        return fig, (ax,)
    return fig, (ax, ax_isim)


def umap_plot(
    c: ClusterAnalysis,
    /,
    title: str | None = None,
    scaling: str = "normalize",
    n_neighbors: int = 15,
    min_dist: float = 0.1,
    metric: str = "euclidean",
    densmap: bool = False,
    workers: int | None = None,
    deterministic: bool = False,
) -> tuple[plt.Figure, tuple[plt.Axes, ...]]:
    r"""Create a UMAP plot from a cluster analysis"""
    color_labels: list[int] = []
    for num, label in zip(c.sizes, c.labels):
        color_labels.extend([label - 1] * num)  # color labels start with 0
    num_top = c.clusters_num
    if workers is None:
        workers = _num_avail_cpus()

    # I don't think these should be transformed, like this, only normalized
    if scaling == "normalize":
        fps_scaled = normalize_features(c.top_unpacked_fps)
    elif scaling == "std":
        scaler = StandardScaler()
        fps_scaled = scaler.fit_transform(c.top_unpacked_fps)
    elif scaling == "none":
        fps_scaled = c.top_unpacked_fps
    else:
        raise ValueError(f"Unknown scaling {scaling}")
    fps_umap = umap.UMAP(
        densmap=densmap,
        random_state=42 if deterministic else None,
        n_components=2,
        n_jobs=workers,
        n_neighbors=n_neighbors,
        min_dist=min_dist,
        metric=metric,
    ).fit_transform(fps_scaled)
    fig, ax = plt.subplots(dpi=250, figsize=(4, 3.5))
    scatter = ax.scatter(
        fps_umap[:, 0],
        fps_umap[:, 1],
        c=color_labels,
        cmap=mpl.colors.ListedColormap(colorcet.glasbey_bw_minc_20[:num_top]),
        edgecolors="none",
        alpha=0.5,
        s=2,
    )
    # t-SNE plots *must be square*
    ax.set_aspect("equal", adjustable="box")
    cbar = plt.colorbar(scatter, label="Cluster label")
    cbar.set_ticks(list(range(num_top)))
    cbar.set_ticklabels(list(map(str, range(1, num_top + 1))))
    ax.set_xlabel("UMAP component 1")
    ax.set_ylabel("UMAP component 2")
    if c.has_all_clusters:
        msg = "UMAP of all clusters"
    else:
        msg = f"UMAP of top {num_top} largest clusters"
    if title is not None:
        msg = f"{msg} for {title}"
    fig.suptitle(msg)
    return fig, (ax,)


def pca_plot(
    c: ClusterAnalysis,
    /,
    title: str | None = None,
    scaling: str = "normalize",
    whiten: bool = False,
) -> tuple[plt.Figure, tuple[plt.Axes, ...]]:
    r"""Create a t-SNE plot from a cluster analysis"""
    color_labels: list[int] = []
    for num, label in zip(c.sizes, c.labels):
        color_labels.extend([label - 1] * num)  # color labels start with 0
    num_top = c.clusters_num

    # I don't think these should be transformed, like this, only normalized
    if scaling == "normalize":
        fps_scaled = normalize_features(c.top_unpacked_fps)
    elif scaling == "std":
        scaler = StandardScaler()
        fps_scaled = scaler.fit_transform(c.top_unpacked_fps)
    elif scaling == "none":
        fps_scaled = c.top_unpacked_fps
    else:
        raise ValueError(f"Unknown scaling {scaling}")
    fps_pca = PCA(n_components=2, whiten=whiten, random_state=1234).fit_transform(
        fps_scaled
    )
    fig, ax = plt.subplots(dpi=250, figsize=(4, 3.5))
    scatter = ax.scatter(
        fps_pca[:, 0],
        fps_pca[:, 1],
        c=color_labels,
        cmap=mpl.colors.ListedColormap(colorcet.glasbey_bw_minc_20[:num_top]),
        edgecolors="none",
        alpha=0.5,
        s=2,
    )
    # t-SNE plots *must be square*
    ax.set_aspect("equal", adjustable="box")
    cbar = plt.colorbar(scatter, label="Cluster label")
    cbar.set_ticks(list(range(num_top)))
    cbar.set_ticklabels(list(map(str, range(1, num_top + 1))))
    ax.set_xlabel("PCA component 1")
    ax.set_ylabel("PCA component 2")
    if c.has_all_clusters:
        msg = "PCA of all clusters"
    else:
        msg = f"PCA of top {num_top} largest clusters"
    if title is not None:
        msg = f"{msg} for {title}"
    fig.suptitle(msg)
    return fig, (ax,)


def tsne_plot(
    c: ClusterAnalysis,
    /,
    title: str | None = None,
    seed: int | None = TSNE_SEED,
    perplexity: int = 30,
    workers: int | None = None,
    scaling: str = "normalize",
    exaggeration: float | None = None,
    do_pca_init: bool = True,
    multiscale: bool = False,
    pca_reduce: int | None = None,
    metric: str = "euclidean",
    dof: float = 1.0,
) -> tuple[plt.Figure, tuple[plt.Axes, ...]]:
    r"""Create a t-SNE plot from a cluster analysis"""
    if workers is None:
        workers = _num_avail_cpus()
    color_labels: list[int] = []
    for num, label in zip(c.sizes, c.labels):
        color_labels.extend([label - 1] * num)  # color labels start with 0
    num_top = c.clusters_num

    # I don't think these should be transformed, like this, only normalized
    if scaling == "normalize":
        fps_scaled = normalize_features(c.top_unpacked_fps)
    elif scaling == "std":
        scaler = StandardScaler()
        fps_scaled = scaler.fit_transform(c.top_unpacked_fps)
    elif scaling == "none":
        fps_scaled = c.top_unpacked_fps
    else:
        raise ValueError(f"Unknown scaling {scaling}")
    if pca_reduce is not None:
        fps_scaled = PCA(n_components=pca_reduce).fit_transform(fps_scaled)

    # Learning rate is set to N / exaggeration (good default)
    # Early exaggeration defaults to max(12, exaggeration) (good default)
    # exaggeration_iter = 250, normal_iter = 500 (good defaults)
    # "pca" is the method used by Dimitry Kovak et. al. (good default), with some jitter
    # added for extra numerical stability
    # Multiscale may help with medium-sized datasets together with downsampling, but
    # it doesn't do much in my tests.
    # NOTE: Dimensionality reduction with PCA to ~50 features seems to mostly preserve
    # cluster structure
    tsne = TSNE(
        n_components=2,
        perplexity=perplexity,
        random_state=seed,
        n_jobs=workers,
        dof=dof,
        exaggeration=exaggeration,  # second-phase exaggeration
        negative_gradient_method="fft",  # faster for large datasets
        initialization="pca" if do_pca_init else "random",
    )
    if multiscale:
        fps_tsne = (
            super(TSNE, tsne)
            .fit(
                fps_scaled,
                affinities=Multiscale(
                    n_jobs=workers,
                    random_state=seed,
                    data=fps_scaled,
                    perplexities=[perplexity, len(fps_scaled) / 100],
                ),
                initialization="pca" if do_pca_init else "random",
            )
            .view(np.ndarray)
        )
    else:
        fps_tsne = tsne.fit_transform(fps_scaled)

    fig, ax = plt.subplots(dpi=250, figsize=(4, 3.5))
    scatter = ax.scatter(
        fps_tsne[:, 0],
        fps_tsne[:, 1],
        c=color_labels,
        cmap=mpl.colors.ListedColormap(colorcet.glasbey_bw_minc_20[:num_top]),
        edgecolors="none",
        alpha=0.5,
        s=2,
    )
    # t-SNE plots *must be square*
    ax.set_aspect("equal", adjustable="box")
    cbar = plt.colorbar(scatter, label="Cluster label")
    cbar.set_ticks(list(range(num_top)))
    cbar.set_ticklabels(list(map(str, range(1, num_top + 1))))
    ax.set_xlabel("t-SNE component 1")
    ax.set_ylabel("t-SNE component 2")
    if c.has_all_clusters:
        msg = "t-SNE of all clusters"
    else:
        msg = f"t-SNE of top {num_top} largest clusters"
    if title is not None:
        msg = f"{msg} for {title}"
    fig.suptitle(msg)
    return fig, (ax,)


def dump_mol_images(
    smiles: tp.Iterable[str],
    clusters: list[list[int]],
    cluster_idx: int = 0,
    batch_size: int = 30,
    limit: int = -1,
) -> None:
    r"""Dump smiles associated with a specific cluster as ``*.png`` image files"""
    if isinstance(smiles, str):
        smiles = [smiles]
    smiles = np.asarray(smiles)
    idxs = clusters[cluster_idx]
    num = 0
    for i, idx_seq in enumerate(batched(idxs, batch_size)):
        if num + len(idx_seq) > limit:
            idx_seq = idx_seq[: num + len(idx_seq) - limit]
        mols = []
        for smi in smiles[list(idx_seq)]:
            mol = Chem.MolFromSmiles(smi)
            if mol is None:
                raise ValueError(f"Could not parse smiles {smi}")
            mols.append(mol)
        img = Draw.MolsToGridImage(mols, molsPerRow=5)
        with open(f"cluster_{cluster_idx}_{i}.png", "wb") as f:
            f.write(img.data)
        num += len(idx_seq)
        if num >= limit:
            break


# For internal use, dispatches a visualization workflow and optionally saves
# plot to disk and/or displays it using mpl
def _dispatch_visualization(
    clusters_path: Path,
    fn_name: str,
    fn: tp.Callable[..., tp.Any],
    fn_kwargs: tp.Any,
    min_size: int = 0,
    smiles: tp.Iterable[str] = (),
    top: int | None = None,
    n_features: int | None = None,
    input_is_packed: bool = True,
    fps_path: Path | None = None,
    title: str | None = None,
    filename: str | None = None,
    verbose: bool = True,
    save: bool = True,
    show: bool = True,
    warn_nan: bool = True,
) -> tuple[plt.Figure, tuple[plt.Axes, ...]]:
    if clusters_path.is_dir():
        clusters_path = clusters_path / "clusters.pkl"
    with open(clusters_path, mode="rb") as f:
        clusters = pickle.load(f)
    if fps_path is None:
        input_fps_path = clusters_path.parent / "input-fps"
        if input_fps_path.is_dir() and _has_files_or_valid_symlinks(input_fps_path):
            fps_path = input_fps_path
        else:
            if fn_name != "summary":
                msg = "Could not find input fingerprints. Please use --fps-path"
                raise RuntimeError(msg)
            else:
                msg = (
                    "Could not find input fingerprints. Please use --fps-path."
                    " Summary plot without fingerprints doesn't include isim values"
                )
                warnings.warn(msg)
    if fps_path is None:
        fps_paths = None
    elif fps_path.is_dir():
        fps_paths = sorted(
            f for f in fps_path.glob("*.npy") if not f.stem.endswith(".indices")
        )
    else:
        fps_paths = [fps_path]
    ca = cluster_analysis(
        clusters,
        fps_paths,
        smiles=smiles,
        top=top,
        n_features=n_features,
        input_is_packed=input_is_packed,
        min_size=min_size,
        warn_nan=warn_nan,
    )
    out = fn(ca, title=title, **fn_kwargs)
    if save:
        if filename is None:
            unique_id = format(random.getrandbits(32), "08x")
            filename = f"{fn_name}-{unique_id}.pdf"
        plt.savefig(Path.cwd() / filename)
    if show:
        plt.show()
    return out



r"""Utilites for manipulating fingerprints and fingerprint files"""

import sys
import math
import weakref
import warnings
import dataclasses
from pathlib import Path
from numpy.typing import NDArray, DTypeLike
import numpy as np
import typing as tp
import multiprocessing as mp
import multiprocessing.shared_memory as shmem

from rich.console import Console
from rdkit.Chem import rdFingerprintGenerator, MolFromSmiles, SanitizeFlags, SanitizeMol

from bblean._config import DEFAULTS
from bblean._console import get_console
from bblean.smiles import _iter_ranges_and_smiles_batches
from bblean.utils import _num_avail_cpus

__all__ = [
    "make_fake_fingerprints",
    "fps_from_smiles",
    "pack_fingerprints",
    "unpack_fingerprints",
]


# Deprecated
def calc_centroid(
    linear_sum: NDArray[np.integer], n_samples: int, *, pack: bool = True
) -> NDArray[np.uint8]:
    warnings.warn(
        "Please use `bblean.similarity.centroid_from_sum(...)` instead",
        DeprecationWarning,
        stacklevel=2,
    )
    if n_samples <= 1:
        centroid = linear_sum.astype(np.uint8, copy=False)
    else:
        centroid = (linear_sum >= n_samples * 0.5).view(np.uint8)
    if pack:
        return np.packbits(centroid, axis=-1)
    return centroid


centroid_from_sum = calc_centroid


def pack_fingerprints(a: NDArray[np.uint8]) -> NDArray[np.uint8]:
    r"""Pack binary (only 0s and 1s) uint8 fingerprint arrays"""
    # packbits may pad with zeros if n_features is not a multiple of 8
    return np.packbits(a, axis=-1)


def unpack_fingerprints(
    a: NDArray[np.uint8], n_features: int | None = None
) -> NDArray[np.uint8]:
    r"""Unpack packed uint8 arrays into binary uint8 arrays (with only 0s and 1s)

    .. note::

        If ``n_features`` is not passed, unpacking will only recover the correct number
        of features if it is a multiple of 8, otherwise fingerprints will be padded with
        zeros to the closest multiple of 8. This is generally not an issue since most
        common fingerprints feature sizes (2048, 1024, etc) are multiples of 8, but if
        you are using a non-standard number of features you should pass ``n_features``
        explicitly.
    """
    # n_features is required to discard padded zeros if it is not a multiple of 8
    return np.unpackbits(a, axis=-1, count=n_features)


def make_fake_fingerprints(
    num: int,
    n_features: int = DEFAULTS.n_features,
    pack: bool = True,
    seed: int | None = None,
    dtype: DTypeLike = np.uint8,
) -> NDArray[np.uint8]:
    r"""Make random fingerprints with statistics similar to (some) real databases"""
    import scipy.stats  # Hide this import since scipy is heavy

    if n_features < 1 or n_features % 8 != 0:
        raise ValueError("n_features must be a multiple of 8, and greater than 0")
    # Generate "synthetic" fingerprints with a popcount distribution
    # similar to one in a real smiles database
    # Fps are guaranteed to *not* be all zeros or all ones
    if pack:
        if np.dtype(dtype) != np.dtype(np.uint8):
            raise ValueError("Only np.uint8 dtype is supported for packed input")
    loc = 750
    scale = 400
    bounds = (0, n_features)
    rng = np.random.default_rng(seed)
    safe_bounds = (bounds[0] + 1, bounds[1] - 1)
    a = (safe_bounds[0] - loc) / scale
    b = (safe_bounds[1] - loc) / scale
    popcounts_fake_float = scipy.stats.truncnorm.rvs(
        a, b, loc=loc, scale=scale, size=num, random_state=rng
    )
    popcounts_fake = np.rint(popcounts_fake_float).astype(np.int64)
    zerocounts_fake = n_features - popcounts_fake
    repeats_fake = np.empty((num * 2), dtype=np.int64)
    repeats_fake[0::2] = popcounts_fake
    repeats_fake[1::2] = zerocounts_fake
    initial = np.tile(np.array([1, 0], np.uint8), num)
    expanded = np.repeat(initial, repeats=repeats_fake)
    fps_fake = rng.permuted(expanded.reshape(num, n_features), axis=-1)
    if pack:
        return np.packbits(fps_fake, axis=1)
    return fps_fake.astype(dtype, copy=False)


def _get_generator(kind: str, n_features: int) -> tp.Any:
    if kind == "rdkit":
        return rdFingerprintGenerator.GetRDKitFPGenerator(fpSize=n_features)
    elif kind == "ecfp4":
        return rdFingerprintGenerator.GetMorganGenerator(radius=2, fpSize=n_features)
    elif kind == "ecfp6":
        return rdFingerprintGenerator.GetMorganGenerator(radius=3, fpSize=n_features)
    elif kind == "topological":
        return rdFingerprintGenerator.GetTopologicalTorsionGenerator(fpSize=n_features)
    elif kind == "ap":
        return rdFingerprintGenerator.GetAtomPairGenerator(fpSize=n_features)
    raise ValueError(f"Unknown kind {kind}. Use 'rdkit|ecfp4|ecfp6|topological|ap'")


def _get_sanitize_flags(sanitize: str) -> tp.Any:
    if sanitize == "all":
        return SanitizeFlags.SANITIZE_ALL
    elif sanitize == "minimal":
        flags = SanitizeFlags.SANITIZE_CLEANUP | SanitizeFlags.SANITIZE_SYMMRINGS
        return flags
    else:
        raise ValueError("Unknown 'sanitize', must be one of 'all', 'minimal'")


@tp.overload
def fps_from_smiles(
    smiles: tp.Iterable[str],
    kind: str = DEFAULTS.fp_kind,
    n_features: int = DEFAULTS.n_features,
    dtype: DTypeLike = np.uint8,
    sanitize: str = "all",
    skip_invalid: tp.Literal[False] = False,
    pack: bool = True,
) -> NDArray[np.uint8]:
    pass


@tp.overload
def fps_from_smiles(
    smiles: tp.Iterable[str],
    kind: str = DEFAULTS.fp_kind,
    n_features: int = DEFAULTS.n_features,
    dtype: DTypeLike = np.uint8,
    sanitize: str = "all",
    skip_invalid: tp.Literal[True] = True,
    pack: bool = True,
) -> tuple[NDArray[np.uint8], NDArray[np.int64]]:
    pass


def fps_from_smiles(
    smiles: tp.Iterable[str],
    kind: str = DEFAULTS.fp_kind,
    n_features: int = DEFAULTS.n_features,
    dtype: DTypeLike = np.uint8,
    sanitize: str = "all",
    skip_invalid: bool = False,
    pack: bool = True,
) -> tp.Union[NDArray[np.uint8], tuple[NDArray[np.uint8], NDArray[np.int64]]]:
    r"""Convert a sequence of smiles into chemical fingerprints"""
    if n_features < 1 or n_features % 8 != 0:
        raise ValueError("n_features must be a multiple of 8, and greater than 0")
    if isinstance(smiles, str):
        smiles = [smiles]

    if pack and not (np.dtype(dtype) == np.dtype(np.uint8)):
        raise ValueError("Packing only supported for uint8 dtype")

    fpg = _get_generator(kind, n_features)

    sanitize_flags = _get_sanitize_flags(sanitize)

    smiles = list(smiles)
    fps = np.empty((len(smiles), n_features), dtype=dtype)

    invalid_idxs = []
    for i, smi in enumerate(smiles):
        mol = MolFromSmiles(smi, sanitize=False)
        if mol is None:
            if skip_invalid:
                invalid_idxs.append(i)
                continue
            else:
                raise ValueError(f"Unable to parse smiles {smi}")
        try:
            SanitizeMol(mol, sanitizeOps=sanitize_flags)
            fps[i, :] = fpg.GetFingerprintAsNumPy(mol)
        except Exception:
            if skip_invalid:
                invalid_idxs.append(i)
                continue
            raise

    if invalid_idxs:
        fps = np.delete(fps, invalid_idxs, axis=0)
    if pack:
        if skip_invalid:
            return pack_fingerprints(fps), np.array(invalid_idxs, dtype=np.int64)
        return pack_fingerprints(fps)
    if skip_invalid:
        return fps, np.array(invalid_idxs, dtype=np.int64)
    return fps


def _get_fps_file_num(path: Path) -> int:
    with open(path, mode="rb") as f:
        major, minor = np.lib.format.read_magic(f)
        shape, _, _ = getattr(np.lib.format, f"read_array_header_{major}_{minor}")(f)
        return shape[0]


def _get_fps_file_shape_and_dtype(
    path: Path, raise_if_invalid: bool = False
) -> tuple[tuple[int, int], np.dtype, bool, bool]:
    with open(path, mode="rb") as f:
        major, minor = np.lib.format.read_magic(f)
        shape, _, dtype = getattr(np.lib.format, f"read_array_header_{major}_{minor}")(
            f
        )
    shape_is_valid = len(shape) == 2
    dtype_is_valid = np.issubdtype(dtype, np.integer)
    if raise_if_invalid and (not shape_is_valid) or (not dtype_is_valid):
        raise ValueError(
            f"Fingerprints file {path} is invalid. Shape: {shape}, DType {dtype}"
        )
    return shape, dtype, shape_is_valid, dtype_is_valid


def _print_fps_file_info(path: Path, console: Console | None = None) -> None:
    if console is None:
        console = Console()
    shape, dtype, shape_is_valid, dtype_is_valid = _get_fps_file_shape_and_dtype(path)

    console.print(f"File: {path.resolve()}")
    has_nonzero = None
    if shape_is_valid and dtype_is_valid:
        console.print("    - [green]Valid fingerprint file[/green]")
        if shape[0] > 0:
            first_fp = np.load(path, mmap_mode="r")[0]
            has_nonzero = (first_fp > 1).any()
            if has_nonzero:
                console.print("    - Guessed format: [cyan]Packed[/cyan]")
            else:
                console.print("    - Guessed format: [magenta]Unpacked[/magenta]")
        else:
            console.print("    - Guessed format: [red]Unknown[/red]")
    else:
        console.print("    - [red]Invalid fingerprint file[/red]")
    if shape_is_valid:
        console.print(f"    - Num. fingerprints: {shape[0]:,}")
        if has_nonzero:
            console.print(
                f"    - Num. features: {shape[1]:,} (guessed unpacked: {shape[1] * 8:,})"  # noqa
            )
        else:
            console.print(f"    - Num. features: {shape[1]:,}")
    else:
        console.print(f"    - Shape: {shape}")
    console.print(f"    - DType: [yellow]{dtype.name}[/yellow]")
    console.print()


class _FingerprintFileSequence:
    def __init__(self, files: tp.Iterable[Path]) -> None:
        self._files = list(files)
        if len(self._files) == 0:
            raise ValueError("At least 1 fingerprint file must be provided")

    def __getitem__(self, idxs: tp.Sequence[int]) -> NDArray[np.uint8]:
        return _get_fingerprints_from_file_seq(self._files, idxs)

    @property
    def shape(self) -> tuple[int, int]:
        shape, dtype, _, _ = _get_fps_file_shape_and_dtype(
            self._files[0], raise_if_invalid=True
        )
        return shape


# TODO: The logic of this function is pretty complicated, maybe there is a way to
# simplify it?
def _get_fingerprints_from_file_seq(
    files: tp.Iterable[Path], idxs: tp.Sequence[int]
) -> NDArray[np.uint8]:
    if sorted(idxs) != list(idxs):
        raise ValueError("idxs must be sorted")
    # Sequence of files is assumed to have indexes in an increasing order,
    # for example, if the first two files have 10k fingerprints, then the
    # assoc. idxs are 0-9999 and 10000-19999. 'idxs' will index this sequence of files
    # iter_idxs = iter(idxs)
    n_features = None
    local_file_idxs = []
    consumed_idxs = 0
    running_count = 0
    for f in files:
        (num, _n_features), _, _, _ = _get_fps_file_shape_and_dtype(
            f, raise_if_invalid=True
        )
        # Fetch idxs Append array([]) if no idxs in the file
        file_idxs = list(
            filter(lambda x: x < running_count + num, idxs[consumed_idxs:])
        )
        consumed_idxs += len(file_idxs)
        local_file_idxs.append(np.array(file_idxs, dtype=np.uint64) - running_count)
        running_count += num

        if n_features is None:
            n_features = _n_features
        elif _n_features != n_features:
            raise ValueError(
                f"Incompatible in fingerprint file {f},"
                f" expected {n_features}, found {_n_features}"
            )
    if len(idxs) != sum(arr.size for arr in local_file_idxs):
        raise ValueError("idxs could not be extracted from files")

    arr = np.empty((len(idxs), tp.cast(int, n_features)), dtype=np.uint8)
    i = 0
    for file, local_idxs in zip(files, local_file_idxs):
        size = local_idxs.size
        if not size:
            continue
        arr[i : i + size] = np.load(file, mmap_mode="r")[local_idxs].astype(
            np.uint8, copy=False
        )
        i += size
    return arr


# TODO: Skipping invalid smiles is a bit inefficient
# NOTE: Mostly convenient for usage in multiprocessing workflows
@dataclasses.dataclass
class _FingerprintFileCreator:
    dtype: str
    out_dir: Path
    out_name: str
    digits: int | None
    pack: bool
    kind: str
    n_features: int
    sanitize: str
    skip_invalid: bool
    verbose: bool

    def __call__(self, input_: tuple[int, tp.Sequence[str]]) -> None:
        console = get_console(self.verbose)
        fpg = _get_generator(self.kind, self.n_features)
        file_idx, batch = input_
        fps = np.empty((len(batch), self.n_features), dtype=self.dtype)
        out_name = self.out_name
        sanitize_flags = _get_sanitize_flags(self.sanitize)
        invalid = []
        for i, smi in enumerate(batch):
            mol = MolFromSmiles(smi, sanitize=False)
            if mol is None:
                if self.skip_invalid:
                    invalid.append(i)
                    continue
                else:
                    raise ValueError(f"Unable to parse smiles {smi}")
            try:
                SanitizeMol(mol, sanitizeOps=sanitize_flags)  # Raises if invalid
            except Exception:
                if self.skip_invalid:
                    invalid.append(i)
                    continue
                raise
            fps[i, :] = fpg.GetFingerprintAsNumPy(mol)
        if self.pack:
            fps = pack_fingerprints(fps)
        if self.digits is not None:
            out_name = f"{out_name}.{str(file_idx).zfill(self.digits)}"
        if invalid:
            prev_num = len(fps)
            fps = np.delete(fps, invalid, axis=0)
            new_num = len(fps)
            console.print(
                f"File {file_idx}: Generated {new_num} fingerprints\n"
                f" File {file_idx}: Skipped {prev_num - new_num} invalid smiles"
            )
        np.save(self.out_dir / out_name, fps)


@dataclasses.dataclass
class _FingerprintArrayFiller:
    invalid_mask_shmem_name: str
    shmem_name: str
    kind: str
    fp_size: int
    pack: bool
    dtype: str
    num_smiles: int
    sanitize: str
    skip_invalid: bool

    def __call__(self, idx_range: tuple[int, int], batch: tp.Sequence[str]) -> None:
        fpg = _get_generator(self.kind, self.fp_size)
        (idx0, idx1) = idx_range
        fps_shmem = shmem.SharedMemory(name=self.shmem_name)
        invalid_mask_shmem = shmem.SharedMemory(name=self.invalid_mask_shmem_name)
        sanitize_flags = _get_sanitize_flags(self.sanitize)

        if self.pack:
            out_dim = (self.fp_size + 7) // 8
        else:
            out_dim = self.fp_size
        fps = np.ndarray(
            (self.num_smiles, out_dim), dtype=self.dtype, buffer=fps_shmem.buf
        )
        invalid_mask = np.ndarray(
            (self.num_smiles,), dtype=np.bool, buffer=invalid_mask_shmem.buf
        )
        for i, smi in zip(range(idx0, idx1), batch):
            mol = MolFromSmiles(smi, sanitize=False)
            if mol is None:
                if self.skip_invalid:
                    invalid_mask[i] = True
                    continue
                else:
                    raise ValueError(f"Unable to parse smiles {smi}")
            try:
                SanitizeMol(mol, sanitizeOps=sanitize_flags)  # Raises if invalid
            except Exception:
                if self.skip_invalid:
                    invalid_mask[i] = True
                    continue
                raise
            fp = fpg.GetFingerprintAsNumPy(mol)
            if self.pack:
                fp = pack_fingerprints(fp)
            fps[i, :] = fp
        fps_shmem.close()
        invalid_mask_shmem.close()


@tp.overload
def fps_from_smiles_parallel(
    smiles: tp.Iterable[str],
    kind: str = DEFAULTS.fp_kind,
    n_features: int = DEFAULTS.n_features,
    dtype: DTypeLike = np.uint8,
    sanitize: str = "all",
    skip_invalid: tp.Literal[False] = False,
    pack: bool = True,
    num_ps: int = 1,
    replace_dummy_atoms: bool = False,
    tab_separated: bool = False,
    mp_context: tp.Any = None,
) -> NDArray[np.uint8]:
    pass


@tp.overload
def fps_from_smiles_parallel(
    smiles: tp.Iterable[str],
    kind: str = DEFAULTS.fp_kind,
    n_features: int = DEFAULTS.n_features,
    dtype: DTypeLike = np.uint8,
    sanitize: str = "all",
    skip_invalid: tp.Literal[True] = True,
    pack: bool = True,
    num_ps: int = 1,
    replace_dummy_atoms: bool = False,
    tab_separated: bool = False,
    mp_context: tp.Any = None,
) -> tp.Union[NDArray[np.uint8], tuple[NDArray[np.uint8], NDArray[np.int64]]]:
    pass


# NOTE: This function is proof of concept and kinda dangerous since it registers
# a custom destructor for the numpy array
# It is also *only usable if called inside an if __name__ == "__main__" guard*
# For now lets hide it
def fps_from_smiles_parallel(
    smiles: tp.Iterable[str],
    kind: str = DEFAULTS.fp_kind,
    n_features: int = DEFAULTS.n_features,
    dtype: DTypeLike = np.uint8,
    sanitize: str = "all",
    skip_invalid: bool = False,
    pack: bool = True,
    num_ps: int | None = None,
    replace_dummy_atoms: bool = False,
    tab_separated: bool = False,
    mp_context: tp.Any = None,
) -> tp.Union[NDArray[np.uint8], tuple[NDArray[np.uint8], NDArray[np.int64]]]:
    r""":meta private:"""
    if mp_context is None:
        mp_context = mp.get_context("forkserver" if sys.platform == "linux" else None)
    if isinstance(smiles, str):
        smiles = [smiles]
    smiles = list(smiles)
    smiles_num = len(smiles)
    if num_ps is None:
        num_ps = _num_avail_cpus()

    if pack:
        out_dim = (n_features + 7) // 8
    else:
        out_dim = n_features
    shmem_size = smiles_num * out_dim * np.dtype(dtype).itemsize
    fps_shmem = shmem.SharedMemory(create=True, size=shmem_size)
    invalid_mask_shmem = shmem.SharedMemory(create=True, size=smiles_num)
    fps_array_filler = _FingerprintArrayFiller(
        shmem_name=fps_shmem.name,
        invalid_mask_shmem_name=invalid_mask_shmem.name,
        kind=kind,
        fp_size=n_features,
        num_smiles=smiles_num,
        dtype=np.dtype(dtype).name,
        pack=pack,
        sanitize=sanitize,
        skip_invalid=skip_invalid,
    )
    num_per_batch = math.ceil(smiles_num / num_ps)
    with mp_context.Pool(processes=num_ps) as pool:
        pool.starmap(
            fps_array_filler,
            _iter_ranges_and_smiles_batches(
                smiles,
                num_per_batch,
                tab_separated,
                replace_dummy_atoms,
                assume_paths=False,
            ),
        )
    fps = np.ndarray((smiles_num, out_dim), dtype=dtype, buffer=fps_shmem.buf)
    mask = np.ndarray((smiles_num,), dtype=np.bool, buffer=invalid_mask_shmem.buf)
    if skip_invalid:
        fps = np.delete(fps, mask, axis=0)
        weakref.finalize(mask, invalid_mask_shmem.close)
        weakref.finalize(mask, invalid_mask_shmem.unlink)
        weakref.finalize(fps, fps_shmem.close)
        weakref.finalize(fps, fps_shmem.unlink)
        return fps, mask

    del mask
    invalid_mask_shmem.close()
    invalid_mask_shmem.unlink()
    weakref.finalize(fps, fps_shmem.close)
    weakref.finalize(fps, fps_shmem.unlink)
    return fps



import typing as tp

import numpy as np
from numpy.typing import NDArray
from inline_snapshot import snapshot

from legacy_merges import (  # type: ignore
    merge_radius,
    merge_diameter,
    merge_tolerance,
)
from bblean import BitBirch
from bblean.merges import (
    NeverMerge,
    RadiusMerge,
    DiameterMerge,
    ToleranceLegacyMerge,
    ToleranceDiameterMerge,
    ToleranceRadiusMerge,
    DiscardSubcluster,
)
from bblean.fingerprints import make_fake_fingerprints
from bblean.similarity import centroid_from_sum

# Cases to test for all merges:
# low|high tolerance


def get_old_and_nom(
    fps: NDArray[np.integer], j: int, case: str = "1, 1"
) -> tuple[NDArray[np.integer], NDArray[np.integer]]:
    if case == "1, 1":
        old = fps[j : j + 1]
        nom = fps[j + 1 : j + 2]
        return old, nom
    if case == "1, >1":
        old = fps[j : j + 1]
        nom = fps[j + 1 : j + 100]
        return old, nom
    if case == ">1, 1":
        old = fps[j : j + 100]
        nom = fps[j + 101 : j + 102]
        return old, nom
    if case == ">1, >1":
        old = fps[j : j + 100]
        nom = fps[j + 101 : j + 200]
        return old, nom
    raise ValueError("Unknown case")


# low|high threshold
# old = 1, nom = 1
# old = 1, nom > 1
# old > 1, nom = 1
# old > 1, nom > 1
def test_non_tolerance() -> None:
    fps = make_fake_fingerprints(
        500, n_features=2048, seed=12620509540149709235, pack=False
    )
    legacy_fns = (
        merge_radius,
        merge_diameter,
    )
    oop_fns = (
        RadiusMerge,
        DiameterMerge,
    )
    thresholds = (0.65, 0.65, 0.3, 0.3)

    for fn_expect, Fn, thresh in zip(legacy_fns, oop_fns, thresholds):
        fn = Fn()
        for case in ("1, 1", "1, >1", ">1, 1", ">1, >1"):
            for j in range(200):
                old, nom = get_old_and_nom(fps, j, case)
                old_ls = old.sum(0)
                nom_ls = nom.sum(0)
                new_ls = old_ls + nom_ls
                old_n = len(old)
                nom_n = len(nom)
                new_n = old_n + nom_n
                cent = centroid_from_sum(new_ls, new_n, pack=False)
                val_expect = fn_expect(
                    thresh, new_ls, cent, new_n, old_ls, nom_ls, old_n, nom_n
                )
                val = fn(thresh, new_ls, new_n, old_ls, nom_ls, old_n, nom_n, [], [])
                assert val == val_expect


# These are designed to trip all cases of tolerance
def test_tolerance_radius() -> None:
    fps = make_fake_fingerprints(
        500, n_features=2048, seed=12620509540149709235, pack=False
    )
    tolerances = (0.00, 1e-8, 0.05, 0.05, 0.9, 0.5)

    expect = [
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        True,
        True,
        True,
        True,
    ]
    idx = 0
    for thresh in (0.23, 1e-3):
        for j, tol in enumerate(tolerances):
            old, nom = get_old_and_nom(fps, j, ">1, >1")
            old_ls = old.sum(0)
            nom_ls = nom.sum(0)
            new_ls = old_ls + nom_ls
            old_n = len(old)
            nom_n = len(nom)
            new_n = old_n + nom_n
            fn = ToleranceRadiusMerge(tolerance=tol)
            val = fn(thresh, new_ls, new_n, old_ls, nom_ls, old_n, nom_n, [], [])
            assert val == expect[idx]
            idx += 1


def test_never_merge() -> None:
    fps = make_fake_fingerprints(
        500, n_features=2048, seed=12620509540149709235, pack=False
    )
    tolerances = range(1, 10)
    for thresh in (0.23, 0.2):
        for j, tol in enumerate(tolerances):
            old, nom = get_old_and_nom(fps, j, ">1, >1")
            old_ls = old.sum(0)
            nom_ls = nom.sum(0)
            new_ls = old_ls + nom_ls
            old_n = len(old)
            nom_n = len(nom)
            new_n = old_n + nom_n
            fn = NeverMerge(tolerance=tol)
            val = fn(thresh, new_ls, new_n, old_ls, nom_ls, old_n, nom_n, [], [])
            assert not val


# These are designed to trip all cases of tolerance
def test_tolerance_diameter() -> None:
    fps = make_fake_fingerprints(
        500, n_features=2048, seed=12620509540149709235, pack=False
    )
    tolerances = (0.00, 1e-8, 0.05, 0.05, 0.9, 0.5)

    expect = [
        False,
        False,
        False,
        False,
        False,
        False,
        False,
        True,
        True,
        True,
        True,
        True,
    ]
    idx = 0
    for thresh in (0.23, 0.2):
        for j, tol in enumerate(tolerances):
            old, nom = get_old_and_nom(fps, j, ">1, >1")
            old_ls = old.sum(0)
            nom_ls = nom.sum(0)
            new_ls = old_ls + nom_ls
            old_n = len(old)
            nom_n = len(nom)
            new_n = old_n + nom_n
            fn = ToleranceDiameterMerge(tolerance=tol)
            val = fn(thresh, new_ls, new_n, old_ls, nom_ls, old_n, nom_n, [], [])
            assert val == expect[idx]
            idx += 1


# These are designed to trip all cases of tolerance
def test_tolerance() -> None:
    fps = make_fake_fingerprints(
        500, n_features=2048, seed=12620509540149709235, pack=False
    )
    legacy_fns = (
        merge_tolerance,
        merge_tolerance,
    )
    oop_fns = (
        ToleranceLegacyMerge,
        ToleranceLegacyMerge,
    )
    thresholds = (0.2, 0.2, 0.2, 0.2)
    tolerances = (0.05, 0.05, 0.90, 0.90)
    for fn_expect, Fn, thresh, tol in zip(legacy_fns, oop_fns, thresholds, tolerances):
        fn = Fn(tol)
        fn._backwards_compat = True  # type: ignore
        for case in ("1, 1", "1, >1", ">1, 1", ">1, >1"):
            for j in range(200):
                old, nom = get_old_and_nom(fps, j, case)
                old_ls = old.sum(0)
                nom_ls = nom.sum(0)
                new_ls = old_ls + nom_ls
                old_n = len(old)
                nom_n = len(nom)
                new_n = old_n + nom_n
                cent = centroid_from_sum(new_ls, new_n, pack=False)
                val_expect = fn_expect(
                    thresh, new_ls, cent, new_n, old_ls, nom_ls, old_n, nom_n, tol
                )
                val = fn(thresh, new_ls, new_n, old_ls, nom_ls, old_n, nom_n, [], [])
                assert val == val_expect


def test_custom_merge() -> None:
    class TrackingDiameterMerge(DiameterMerge):
        def __init__(self, max_cluster_size: int) -> None:
            self.max_cluster_size = max_cluster_size
            self.redundant_mol_idxs: list[int] = []

        def on_check_merge_end(
            self,
            accepted: bool,
            old_idxs: tp.Sequence[int],
            nominee_idxs: tp.Sequence[int],
        ) -> None:
            if accepted and len(old_idxs) >= self.max_cluster_size:
                self.redundant_mol_idxs.extend(nominee_idxs)
                raise DiscardSubcluster

    merge_fn = TrackingDiameterMerge(max_cluster_size=32)

    fps = make_fake_fingerprints(
        1000, n_features=2048, seed=12620509540149709235, pack=False
    )
    tree = BitBirch(threshold=0.3, merge_criterion=merge_fn)
    tree.fit(fps)
    assert merge_fn.redundant_mol_idxs == snapshot(
        [
            610,
            611,
            612,
            613,
            614,
            615,
            616,
            617,
            618,
            620,
            621,
            622,
            623,
            625,
            627,
            628,
            629,
            646,
            658,
            679,
            681,
            682,
            685,
            688,
            695,
            696,
            708,
            713,
            714,
            716,
            719,
            720,
            722,
            733,
            734,
            735,
            737,
            740,
            741,
            742,
            743,
            744,
            745,
            746,
            747,
            748,
            749,
            750,
            751,
            752,
            753,
            756,
            757,
            758,
            759,
            761,
            765,
            766,
            767,
            774,
            775,
            777,
            778,
            781,
            786,
            790,
            794,
            795,
            797,
        ]
    )
    assert sum(tree.get_assignments() == 0) == snapshot(69)



from numpy.typing import NDArray
import numpy as np
import pytest
from inline_snapshot import snapshot

from bblean.sklearn import BitBirch, UnpackedBitBirch
from bblean.fingerprints import make_fake_fingerprints


@pytest.fixture
def fps() -> NDArray[np.integer]:
    return make_fake_fingerprints(
        100, n_features=2048, seed=12620509540149709235, pack=True
    )


@pytest.fixture
def unpacked_fps() -> NDArray[np.integer]:
    return make_fake_fingerprints(
        100, n_features=2048, seed=12620509540149709235, pack=False
    )


def test_packed_bitbirch_fit(fps: NDArray[np.integer]) -> None:
    tree = BitBirch(branching_factor=50, threshold=0.1, merge_criterion="diameter")
    tree.fit(fps)
    assert tree.labels_.tolist()[:5] == snapshot([1, 1, 1, 1, 1])
    out = tree.transform(fps)
    assert out.tolist()[:5] == snapshot(
        [
            [0.9953650057937428],
            [0.9955357142857143],
            [0.9965928449744463],
            [0.9977477477477478],
            [0.9977553310886644],
        ]
    )
    tree.set_output(transform="pandas")
    out_df = tree.transform(fps)
    assert out_df.values.tolist()[:5] == snapshot(
        [
            [0.9953650057937428],
            [0.9955357142857143],
            [0.9965928449744463],
            [0.9977477477477478],
            [0.9977553310886644],
        ]
    )
    assert tree.get_params() == snapshot(
        {
            "branching_factor": 50,
            "compute_labels": True,
            "merge_criterion": "diameter",
            "threshold": 0.1,
            "tolerance": None,
        }
    )
    tree.set_params(threshold=0.65)
    assert tree.threshold == 0.65
    out_feats = tree.get_feature_names_out()
    for i, f in enumerate(out_feats):
        assert f == f"bitbirch{i}"


def test_packed_bitbirch_fit_transform(fps: NDArray[np.integer]) -> None:
    tree = BitBirch(branching_factor=50, threshold=0.1, merge_criterion="diameter")
    out = tree.fit_transform(fps)
    assert out.tolist()[:5] == snapshot(
        [
            [0.9953650057937428],
            [0.9955357142857143],
            [0.9965928449744463],
            [0.9977477477477478],
            [0.9977553310886644],
        ]
    )


def test_packed_bitbirch_predict(fps: NDArray[np.integer]) -> None:
    tree = BitBirch(branching_factor=50, threshold=0.1, merge_criterion="diameter")
    out = tree.fit_predict(fps)
    assert out.tolist()[:5] == snapshot([1, 1, 1, 1, 1])
    fps2 = make_fake_fingerprints(
        500, n_features=2048, seed=16297711681646304030, pack=True
    )
    out = tree.predict(fps2)
    assert out.tolist()[:5] == snapshot([1, 1, 1, 1, 1])
    tree.partial_fit(fps2)
    assert tree.labels_.tolist()[:5] == snapshot([1, 1, 1, 1, 1])


def test_unpacked_bitbirch_fit(unpacked_fps: NDArray[np.integer]) -> None:
    tree = UnpackedBitBirch(
        branching_factor=50, threshold=0.1, merge_criterion="diameter"
    )
    tree.fit(unpacked_fps)
    assert tree.labels_.tolist()[:5] == snapshot([1, 1, 1, 1, 1])
    out = tree.transform(unpacked_fps)
    assert out.tolist()[:5] == snapshot(
        [
            [0.9953650057937428],
            [0.9955357142857143],
            [0.9965928449744463],
            [0.9977477477477478],
            [0.9977553310886644],
        ]
    )
    tree.set_output(transform="pandas")
    out_df = tree.transform(unpacked_fps)
    assert out_df.values.tolist()[:5] == snapshot(
        [
            [0.9953650057937428],
            [0.9955357142857143],
            [0.9965928449744463],
            [0.9977477477477478],
            [0.9977553310886644],
        ]
    )
    assert tree.get_params() == snapshot(
        {
            "branching_factor": 50,
            "compute_labels": True,
            "merge_criterion": "diameter",
            "threshold": 0.1,
            "tolerance": None,
        }
    )
    tree.set_params(threshold=0.65)
    assert tree.threshold == 0.65
    out_feats = tree.get_feature_names_out()
    for i, f in enumerate(out_feats):
        assert f == f"unpackedbitbirch{i}"


def test_unpacked_bitbirch_fit_transform(unpacked_fps: NDArray[np.integer]) -> None:
    tree = UnpackedBitBirch(
        branching_factor=50, threshold=0.1, merge_criterion="diameter"
    )
    out = tree.fit_transform(unpacked_fps)
    assert out.tolist()[:5] == snapshot(
        [
            [0.9953650057937428],
            [0.9955357142857143],
            [0.9965928449744463],
            [0.9977477477477478],
            [0.9977553310886644],
        ]
    )


def test_unpacked_bitbirch_predict(unpacked_fps: NDArray[np.integer]) -> None:
    tree = UnpackedBitBirch(
        branching_factor=50, threshold=0.1, merge_criterion="diameter"
    )
    out = tree.fit_predict(unpacked_fps)
    assert out.tolist()[:5] == snapshot([1, 1, 1, 1, 1])
    fps2 = make_fake_fingerprints(
        500, n_features=2048, seed=16297711681646304030, pack=False
    )
    out = tree.predict(fps2)
    assert out.tolist()[:5] == snapshot([1, 1, 1, 1, 1])
    tree.partial_fit(fps2)
    assert tree.labels_.tolist()[:5] == snapshot([1, 1, 1, 1, 1])



r"""Global defaults and related utilities"""

from copy import deepcopy
from pathlib import Path
import typing as tp
import json
import sys
import dataclasses
import multiprocessing as mp
import os

import numpy as np

from bblean._memory import system_mem_gib
from bblean.utils import (
    _cpu_name,
    cpp_extensions_are_enabled,
    cpp_extensions_are_installed,
)


@dataclasses.dataclass(slots=True)
class BitBirchConfig:
    threshold: float = 0.30
    branching_factor: int = 254
    merge_criterion: str = "diameter"
    refine_merge_criterion: str = "tolerance-diameter"
    refine_threshold_change: float = 0.0
    tolerance: float = 0.05
    n_features: int = 2048
    fp_kind: str = "ecfp4"


DEFAULTS = BitBirchConfig()

TSNE_SEED = 42


def collect_system_specs_and_dump_config(
    config: dict[str, tp.Any],
) -> None:
    config = deepcopy(config)
    config_path = Path(config["out_dir"]) / "config.json"
    total_mem, avail_mem = system_mem_gib()
    # System info
    config["cpp_extensions_enabled"] = cpp_extensions_are_enabled()
    config["cpp_extensions_installed"] = cpp_extensions_are_installed()
    config["total_memory_gib"] = total_mem
    config["initial_available_memory_gib"] = avail_mem
    config["platform"] = sys.platform
    config["cpu"] = _cpu_name()
    config["numpy_version"] = np.__version__
    config["python_version"] = sys.version.split()[0]
    # Multiprocessing info
    if config.get("num_processes", 1) > 1:
        config["multiprocessing_start_method"] = mp.get_start_method()
        config["visible_cpu_cores"] = os.cpu_count()

    # Dump config after checking if the output dir has files
    with open(config_path, mode="wt", encoding="utf-8") as f:
        json.dump(config, f, indent=4)



import pytest

from bblean.utils import min_safe_uint


def test_min_safe_uint() -> None:
    assert min_safe_uint(254) == "uint8"
    assert min_safe_uint(255) == "uint8"
    assert min_safe_uint(256) == "uint16"
    assert min_safe_uint(65534) == "uint16"
    assert min_safe_uint(65535) == "uint16"
    assert min_safe_uint(65536) == "uint32"
    assert min_safe_uint(4294967295) == "uint32"
    assert min_safe_uint(4294967296) == "uint64"
    assert min_safe_uint(18446744073709551615) == "uint64"
    with pytest.raises(ValueError):
        min_safe_uint(18446744073709551616)
